classnames = [
    "red and white circle 20 kph speed limit",
    "red and white circle 30 kph speed limit",
    "red and white circle 50 kph speed limit",
    "red and white circle 60 kph speed limit",
    "red and white circle 70 kph speed limit",
    "red and white circle 80 kph speed limit",
    "end / de-restriction of 80 kph speed limit",
    "red and white circle 100 kph speed limit",
    "red and white circle 120 kph speed limit",
    "red and white circle red car and black car no passing",
    "red and white circle red truck and black car no passing",
    "red and white triangle road intersection warning",
    "white and yellow diamond priority road",
    "red and white upside down triangle yield right-of-way",
    "stop",
    "empty red and white circle",
    "red and white circle no truck entry",
    "red circle with white horizonal stripe no entry",
    "red and white triangle with exclamation mark warning",
    "red and white triangle with black left curve approaching warning",
    "red and white triangle with black right curve approaching warning",
    "red and white triangle with black double curve approaching warning",
    "red and white triangle rough / bumpy road warning",
    "red and white triangle car skidding / slipping warning",
    "red and white triangle with merging / narrow lanes warning",
    "red and white triangle with person digging / construction / road work warning",
    "red and white triangle with traffic light approaching warning",
    "red and white triangle with person walking warning",
    "red and white triangle with child and person walking warning",
    "red and white triangle with bicyle warning",
    "red and white triangle with snowflake / ice warning",
    "red and white triangle with deer warning",
    "white circle with gray strike bar no speed limit",
    "blue circle with white right turn arrow mandatory",
    "blue circle with white left turn arrow mandatory",
    "blue circle with white forward arrow mandatory",
    "blue circle with white forward or right turn arrow mandatory",
    "blue circle with white forward or left turn arrow mandatory",
    "blue circle with white keep right arrow mandatory",
    "blue circle with white keep left arrow mandatory",
    "blue circle with white arrows indicating a traffic circle",
    "white circle with gray strike bar indicating no passing for cars has ended",
    "white circle with gray strike bar indicating no passing for trucks has ended",
]

templates = [
    lambda c: f'a zoomed in photo of a "{c}" traffic sign.',
    lambda c: f'a centered photo of a "{c}" traffic sign.',
    lambda c: f'a close up photo of a "{c}" traffic sign.',
]



import logging
import os
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Optional, TypeAlias, Union, cast  # noqa: F401

from omegaconf import DictConfig, flag_override
from transformers import PreTrainedModel, PreTrainedTokenizer
from typing_extensions import override

from fusion_bench.modelpool import BaseModelPool
from fusion_bench.utils import instantiate
from fusion_bench.utils.dtype import parse_dtype

if TYPE_CHECKING:
    from transformers import LlamaForSequenceClassification

log = logging.getLogger(__name__)


class SequenceClassificationModelPool(BaseModelPool):

    def __init__(
        self,
        models,
        *,
        tokenizer: Optional[DictConfig],
        model_kwargs: Optional[DictConfig] = None,
        **kwargs,
    ):
        super().__init__(models, **kwargs)
        # process `model_kwargs`
        self._tokenizer = tokenizer
        self._model_kwargs = model_kwargs
        if self._model_kwargs is None:
            self._model_kwargs = DictConfig({})
        with flag_override(self._model_kwargs, "allow_objects", True):
            if hasattr(self._model_kwargs, "torch_dtype"):
                self._model_kwargs.torch_dtype = parse_dtype(
                    self._model_kwargs.torch_dtype
                )

    @override
    def load_model(
        self,
        model_name_or_config: str | DictConfig,
        *args,
        **kwargs,
    ) -> Union[PreTrainedModel, "LlamaForSequenceClassification"]:
        model_kwargs = deepcopy(self._model_kwargs)
        model_kwargs.update(kwargs)
        if isinstance(model_name_or_config, str):
            log.info(f"Loading model: {model_name_or_config}", stacklevel=2)
        return super().load_model(model_name_or_config, *args, **model_kwargs)

    def load_tokenizer(self, *args, **kwargs) -> PreTrainedTokenizer:
        assert self._tokenizer is not None, "Tokenizer is not defined in the config"
        log.info("Loading tokenizer.", stacklevel=2)
        tokenizer = instantiate(self._tokenizer, *args, **kwargs)
        return tokenizer

    @override
    def save_model(
        self,
        model: PreTrainedModel,
        path: str,
        push_to_hub: bool = False,
        model_dtype: Optional[str] = None,
        save_tokenizer: bool = False,
        tokenizer_kwargs=None,
        **kwargs,
    ):
        """
        Save the model to the specified path.

        Args:
            model (PreTrainedModel): The model to be saved.
            path (str): The path where the model will be saved.
            push_to_hub (bool, optional): Whether to push the model to the Hugging Face Hub. Defaults to False.
            save_tokenizer (bool, optional): Whether to save the tokenizer along with the model. Defaults to False.
            **kwargs: Additional keyword arguments passed to the `save_pretrained` method.
        """
        path = os.path.expanduser(path)
        if save_tokenizer:
            if tokenizer_kwargs is None:
                tokenizer_kwargs = {}
            # load the tokenizer
            tokenizer = self.load_tokenizer(**tokenizer_kwargs)
            tokenizer.save_pretrained(
                path,
                push_to_hub=push_to_hub,
            )
        if model_dtype is not None:
            model.to(dtype=parse_dtype(model_dtype))
        model.save_pretrained(
            path,
            push_to_hub=push_to_hub,
            **kwargs,
        )



import logging
import os
from copy import deepcopy
from functools import cache
from typing import Dict, List, cast

import lightning as L
import torch
from omegaconf import DictConfig
from torch import Tensor, nn
from torch.nn.modules import Module
from torch.utils.data import DataLoader
from tqdm.autonotebook import tqdm
from transformers import GPT2ForSequenceClassification, GPT2Model
from transformers.data import default_data_collator
from transformers.models.gpt2.modeling_gpt2 import Conv1D

from fusion_bench.mixins import LightningFabricMixin, auto_register_config
from fusion_bench.utils import timeit_context

from .regmean import RegMeanAlgorithm

log = logging.getLogger(__name__)


@auto_register_config
class RegMeanAlgorithmForGPT2(
    LightningFabricMixin,
    RegMeanAlgorithm,
):
    _include_module_type = [Conv1D]
    classifiers = {}

    def __init__(self, cache_dir: str, batch_size: int, num_workers: int, **kwargs):
        super().__init__(**kwargs)

    def on_regmean_start(self):
        for model_name in self.modelpool.model_names:
            classifier = cast(
                GPT2ForSequenceClassification,
                self.modelpool.load_classifier(model_name),
            ).requires_grad_(False)
            classifier.transformer = None
            classifier = classifier.to(self.fabric.device)
            self.classifiers[model_name] = classifier

    def compute_logits(self, module: GPT2Model, batch, task: str) -> Tensor:
        self.classifiers[task].transformer = module
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]

        outputs = self.classifiers[task](input_ids, attention_mask=attention_mask)
        logits = outputs.logits
        assert logits.dim() == 2
        return logits

    def get_regmean_weights(
        self,
        model_name: str,
        model: Module,
        train_dataset,
        linear_modules_to_merge: Dict[str, Module],
    ):
        # setup dataloader
        train_dataloader = DataLoader(
            train_dataset,
            batch_size=self.config.batch_size,
            shuffle=True,
            num_workers=self.config.num_workers,
            collate_fn=default_data_collator,
            pin_memory=True,
        )
        train_dataloader = self.fabric.setup_dataloaders(train_dataloader)
        model = self.fabric.setup(model)

        def compute_regmean_weights(module_name: str):
            """
            compute the regmean weights, a hook function to deal with each module's input
            :param module_name: str, module name
            :return:
            """

            def hook(module: nn.Module, input: tuple, output: torch.Tensor):
                # Tensor, shape (batch_size, sequence_length, hidden_dim)
                x = cast(Tensor, input[0]).detach()
                batch_num_actual_examples = x.shape[0]
                # Tensor, shape (batch_size * sequence_length, hidden_dim)
                x = x.reshape(-1, x.shape[-1])
                # Tensor, shape (hidden_dim, hidden_dim)
                xtx = torch.matmul(x.transpose(0, 1), x)
                # store the averaged weights in regmean_weights
                if module_name not in regmean_weights.keys():
                    regmean_weights[module_name] = xtx / x.shape[0]
                    num_computed_examples[module_name] = x.shape[0]
                    num_actual_examples[module_name] = batch_num_actual_examples
                else:
                    regmean_weights[module_name] = (
                        regmean_weights[module_name]
                        * num_computed_examples[module_name]
                        + xtx
                    ) / (num_computed_examples[module_name] + x.shape[0])
                    num_computed_examples[module_name] += x.shape[0]
                    num_actual_examples[module_name] += batch_num_actual_examples

            return hook

        handles = []
        # dictionary, regmean matrices for each linear module inputs
        regmean_weights = {}
        # dictionary, number of examples (multiplied the sequence length) used for computing regmean matrices
        num_computed_examples = {}
        # dictionary, number of actual examples used for computing regmean matrices
        num_actual_examples = {}

        for module_name, linear_module_to_merge in linear_modules_to_merge.items():
            # register a hook in the forward process
            handle = linear_module_to_merge.register_forward_hook(
                compute_regmean_weights(module_name=module_name)
            )
            handles.append(handle)
        for step, batch in tqdm(
            enumerate(train_dataloader),
            desc=f"computing regmean weights for model {model_name}",
        ):
            if (
                len(num_actual_examples) > 0
                and list(num_actual_examples.values())[0]
                >= self.config.num_regmean_examples
            ):
                break
            logits = self.compute_logits(model, batch, model_name)

        # remove the added hook
        for handle in handles:
            handle.remove()

        for module_name in regmean_weights.keys():
            regmean_weights[module_name] = regmean_weights[module_name].detach().cpu()

        return regmean_weights



"""
Continual Model Merging without Data: Dual Projections for Balancing Stability and Plasticity. NeurIPS, 2025.


Example:

fusion_bench \
    method=dop/dop \
    modelpool=CLIPVisionModelPool/clip-vit-base-patch32_TA8_model_only \
    taskpool=CLIPVisionModelTaskPool/clip-vit-classification_TA8
"""

import logging
import os
import random
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, cast

import lightning as L
import numpy as np
import torch
from omegaconf import DictConfig
from torch import Tensor, nn
from torch.autograd import Variable
from tqdm.auto import tqdm
from transformers import CLIPVisionModel

from fusion_bench import BaseAlgorithm, BaseModelPool, auto_register_config
from fusion_bench.method.simple_average import simple_average
from fusion_bench.mixins import LightningFabricMixin
from fusion_bench.taskpool import CLIPVisionModelTaskPool
from fusion_bench.utils import seed_everything_by_time
from fusion_bench.utils.json import save_to_json

from .min_norm_solvers import MinNormSolver, gradient_normalizers
from .utils import is_leaf_module, svd

log = logging.getLogger(__name__)


@auto_register_config
class ContinualDOPForCLIP(BaseAlgorithm, LightningFabricMixin):

    def __init__(
        self,
        seed: Optional[int] = None,
        shuffle_order: bool = False,
        save_on_every_step: bool = True,
        evaluate_on_every_step: bool = False,
        lr: float = 1e-4,
        num_steps: int = 200,
        mgda: bool = True,
        ema: bool = True,
        ema_beta: float = 0.99,
        alpha: float = None,
        svd_epsilon: float = 1.0,
        svd_proj_space: str = "uv",
        **kwargs,
    ):
        self.lr = lr
        self.num_steps = num_steps
        self.mgda = mgda
        self.ema = ema
        self.ema_beta = ema_beta
        self.alpha = alpha
        self.svd_epsilon = svd_epsilon
        self.svd_proj_space = svd_proj_space
        self.seed = seed
        self.shuffle_order = shuffle_order
        self.save_on_every_step = save_on_every_step
        self.evaluate_on_every_step = evaluate_on_every_step

        assert (
            self.svd_epsilon >= 0 and self.svd_epsilon <= 1
        ), "The svd_epsilon should be in the range of [0, 1]"
        assert (
            self.alpha >= 0 and self.alpha <= 1
        ), "The alpha should be in the range of [0, 1]"
        super().__init__(**kwargs)

    def print_params(self, pretrained_model):
        total_params = 0
        linear_params = 0
        linear_weight_params = 0
        for module_name, module in pretrained_model.named_modules():
            if not is_leaf_module(module):
                continue
            if isinstance(module, nn.Linear):
                linear_params += sum(p.numel() for n, p in module.named_parameters())
                linear_weight_params += sum(
                    p.numel() for n, p in module.named_parameters() if "weight" in n
                )
            total_params += sum(p.numel() for p in module.parameters())

        linear_ratio = linear_params / total_params * 100
        linear_weight_ratio = linear_weight_params / total_params * 100
        print(f"Total Parameters: {total_params}")
        print(f"Linear Parameters: {linear_params}")
        print(f"Linear Weight Parameters: {linear_weight_params}")
        print(f"Linear Ratio: {linear_ratio:.2f}%")
        print(f"Linear Weight Ratio: {linear_weight_ratio:.2f}%")

    def run(self, modelpool: BaseModelPool):
        if self.seed is not None:
            L.seed_everything(self.seed)
        else:
            seed_everything_by_time(self.fabric)

        # get the model names, shuffle if needed
        # the model names will be saved to the log directory as `model_names.json`
        model_names = modelpool.model_names
        if self.shuffle_order:
            random.shuffle(model_names)
        if self.log_dir is not None:
            save_to_json(model_names, os.path.join(self.log_dir, "model_names.json"))

        if self.evaluate_on_every_step:
            """Configuration for the test datasets"""
            self.taskpool = cast(CLIPVisionModelTaskPool, self._program.taskpool)
            self._test_datasets = deepcopy(self.taskpool._test_datasets)

        pretrained_model = modelpool.load_pretrained_model()

        merged_model = None
        for model_idx, model_name in enumerate(model_names):
            print(
                f"--------- Optimizing {model_idx + 1}/{len(model_names)}-th with {model_name} ---------"
            )
            if model_idx == 0:
                merged_model = modelpool.load_model(model_names[0])
            else:
                merged_model = self._layer_wise_optimize(
                    model_names=["merged", model_name],
                    pretrained_model=deepcopy(pretrained_model),
                    finetuned_models={
                        "merged": merged_model,
                        model_name: modelpool.load_model(model_name),
                    },
                    model_idx=model_idx,
                )

            if self.save_on_every_step:
                self.save_merged_model(merged_model, model_idx)

            if self.evaluate_on_every_step:
                self.taskpool._is_setup = False
                self.taskpool._test_datasets = DictConfig(
                    {n: self._test_datasets[n] for n in model_names[: model_idx + 1]}
                )
                report = self.taskpool.evaluate(deepcopy(merged_model))
                save_to_json(report, Path(self.log_dir) / f"report_{model_idx}.json")

        return merged_model

    def _layer_wise_optimize(
        self,
        model_names: List[str],
        pretrained_model: nn.Module,
        finetuned_models: Dict[str, nn.Module],
        model_idx: int,
    ):
        time_cost = []
        for module_name, module in pretrained_model.named_modules():
            if not is_leaf_module(module):
                continue

            if isinstance(module, nn.Linear):
                if module.weight.requires_grad:
                    import time

                    start_time = time.time()
                    merged_weight = self._optimize_weight(
                        module.weight,
                        {
                            model_name: finetuned_models[model_name]
                            .get_submodule(module_name)
                            .weight
                            for model_name in model_names
                        },
                        module_name,
                        model_idx,
                    )
                    end_time = time.time()
                    time_cost.append(end_time - start_time)
                    module.weight.data = merged_weight.data
                else:
                    module.weight.data = simple_average(
                        [
                            finetuned_models[model_name]
                            .get_submodule(module_name)
                            .weight
                            for model_name in model_names
                        ]
                    )
                if module.bias is not None:
                    module.bias.data = simple_average(
                        [
                            finetuned_models[model_name].get_submodule(module_name).bias
                            for model_name in model_names
                        ]
                    )
            else:
                simple_average(
                    [
                        finetuned_models[model_name].get_submodule(module_name)
                        for model_name in model_names
                    ],
                    base_module=module,
                )

        return pretrained_model

    def _optimize_weight(
        self,
        pretrained_weight: Tensor,
        finetuned_weights: Dict[str, Tensor],
        module_name: str,
        model_idx: int,
    ):
        assert (
            self.fabric.world_size == 1
        ), "This algorithm is not currently supported in distributed training"

        pretrained_weight = self.fabric.to_device(pretrained_weight.detach())
        finetuned_weights = {
            model_name: self.fabric.to_device(finetuned_weight.detach())
            for model_name, finetuned_weight in finetuned_weights.items()
        }

        merged_weight = self.fabric.to_device(
            nn.Parameter(
                simple_average(
                    [
                        finetuned_weight.detach()
                        for finetuned_weight in finetuned_weights.values()
                    ]
                ),
                requires_grad=True,
            )
        )

        # Compute SVD of the difference between the finetuned and pretrained weights
        proj_u_dict = {}
        proj_v_dict = {}
        proj_s_dict = {}
        for i, finetuned_weight in enumerate(finetuned_weights.values()):
            finetuned_tv = finetuned_weight - pretrained_weight
            u, s, v = svd(finetuned_tv, full_matrices=True)
            epsilon = 1.0 if self.svd_epsilon > 1.0 else self.svd_epsilon
            cumsum_ratio = s.cumsum(dim=0) / s.sum()
            split_rank = torch.searchsorted(cumsum_ratio, epsilon).item()
            u_main = u[:, :split_rank]
            v_main = v[:, :split_rank]
            s_main = s[:split_rank]
            proj_u_dict[i] = u_main
            proj_v_dict[i] = v_main
            proj_s_dict[i] = s_main

        if self.mgda:
            if self.ema:
                ema_sol = [self.alpha, 1 - self.alpha]
            # This is multiple-gradient descent algorithm (MGDA) optimization
            optimizer = torch.optim.Adam([merged_weight], lr=self.lr)
            all_losses = [[], []]
            all_alphas = [[], []]
            for step_idx in tqdm(
                range(self.num_steps), desc=f"Optimizing {module_name} weight"
            ):
                # Scaling the loss functions based on the algorithm choice
                loss_data = {}
                grads = {}
                for i, finetuned_weight in enumerate(finetuned_weights.values()):
                    proj_u = proj_u_dict[i]
                    proj_v = proj_v_dict[i]
                    proj_s = proj_s_dict[i]
                    delta_tv = merged_weight - finetuned_weight
                    loss_i = self.cal_loss_i(delta_tv, proj_s, proj_u, proj_v)
                    loss_data[i] = float(loss_i.data)

                    all_losses[i].append(float(loss_i.data))

                    optimizer.zero_grad()
                    loss_i.backward()
                    grads[i] = Variable(
                        merged_weight.grad.data.clone(), requires_grad=False
                    )

                # Normalize all gradients
                gn = gradient_normalizers(
                    grads=grads, losses=loss_data, normalization_type="loss"
                )
                for i, _ in enumerate(finetuned_weights.values()):
                    grads[i] = grads[i] / float(gn[i])

                # Frank-Wolfe iteration to compute scales.
                sol, min_norm = MinNormSolver.find_min_norm_element(
                    [[grads[i]] for i in range(len(finetuned_weights.values()))]
                )

                if self.ema:
                    ema_sol = [
                        self.ema_beta * ema_sol[i] + (1 - self.ema_beta) * float(sol[i])
                        for i in range(len(sol))
                    ]
                    sol = ema_sol
                    all_alphas[0].append(ema_sol[0])
                    all_alphas[1].append(ema_sol[1])

                # Scaled back-propagation
                loss = 0
                for i, finetuned_weight in enumerate(finetuned_weights.values()):
                    # Comptue gradients of each loss function wrt parameters
                    proj_u = proj_u_dict[i]
                    proj_v = proj_v_dict[i]
                    proj_s = proj_s_dict[i]
                    delta_tv = merged_weight - finetuned_weight
                    loss_i = self.cal_loss_i(delta_tv, proj_s, proj_u, proj_v)
                    loss += float(sol[i]) * loss_i

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

        else:
            # This is a naive weighted optimization
            optimizer = torch.optim.Adam([merged_weight], lr=self.lr)
            for step_idx in tqdm(
                range(self.num_steps), desc=f"Optimizing {module_name} weight"
            ):
                loss = 0
                for i, finetuned_weight in enumerate(finetuned_weights.values()):
                    proj_u = proj_u_dict[i]
                    proj_v = proj_v_dict[i]
                    proj_s = proj_s_dict[i]
                    delta_tv = merged_weight - finetuned_weight
                    loss_i = self.cal_loss_i(delta_tv, proj_s, proj_u, proj_v)
                    loss += self.alpha * loss_i if i == 0 else (1 - self.alpha) * loss_i

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

        return merged_weight.detach().cpu()

    def cal_loss_i(self, delta_tv, proj_s, proj_u, proj_v):
        proj_delta_1 = torch.diag(proj_s) @ proj_u.T @ delta_tv
        proj_delta_2 = delta_tv @ proj_v @ torch.diag(proj_s)
        loss_i_u = torch.linalg.matrix_norm(proj_delta_1, ord="fro") ** 2
        loss_i_v = torch.linalg.matrix_norm(proj_delta_2, ord="fro") ** 2
        if self.svd_proj_space == "uv":
            loss_i = loss_i_u + loss_i_v
        elif self.svd_proj_space == "u":
            loss_i = loss_i_u
        elif self.svd_proj_space == "v":
            loss_i = loss_i_v
        else:
            raise ValueError("Invalid svd_proj_space")

        return loss_i

    def save_merged_model(self, merged_model: CLIPVisionModel, step: int):
        os.makedirs(Path(self.log_dir) / "checkpoints", exist_ok=True)
        merged_model.save_pretrained(
            Path(self.log_dir) / "checkpoints" / f"merged_model_{step}"
        )



import logging
import time

log = logging.getLogger(__name__)


class timeit_context:
    """
    A context manager for measuring and logging execution time of code blocks.

    This context manager provides precise timing measurements with automatic logging
    of elapsed time. It supports nested timing contexts with proper indentation
    for hierarchical timing analysis, making it ideal for profiling complex
    operations with multiple sub-components.

    Args:
        msg (str, optional): Custom message to identify the timed code block.
            If provided, logs "[BEGIN] {msg}" at start and includes context
            in the final timing report. Defaults to None.
        loglevel (int, optional): Python logging level for output messages.
            Uses standard logging levels (DEBUG=10, INFO=20, WARNING=30, etc.).
            Defaults to logging.INFO.

    Example:
        Basic usage:
        ```python
        with timeit_context("data loading"):
            data = load_large_dataset()
        # Logs: [BEGIN] data loading
        # Logs: [END]   Elapsed time: 2.34s
        ```

        Nested timing:
        ```python
        with timeit_context("model training"):
            with timeit_context("data preprocessing"):
                preprocess_data()
            with timeit_context("forward pass"):
                model(data)
        # Output shows nested structure:
        # [BEGIN] model training
        #   [BEGIN] data preprocessing
        #   [END]   Elapsed time: 0.15s
        #   [BEGIN] forward pass
        #   [END]   Elapsed time: 0.89s
        # [END]   Elapsed time: 1.04s
        ```

        Custom log level:
        ```python
        with timeit_context("debug operation", loglevel=logging.DEBUG):
            debug_function()
        ```
    """

    nest_level = -1

    def _log(self, msg):
        """
        Internal method for logging messages with appropriate stack level.

        This helper method ensures that log messages appear to originate from
        the caller's code rather than from internal timer methods, providing
        more useful debugging information.

        Args:
            msg (str): The message to log at the configured log level.
        """
        log.log(self.loglevel, msg, stacklevel=3)

    def __init__(self, msg: str = None, loglevel=logging.INFO) -> None:
        """
        Initialize a new timing context with optional message and log level.

        Args:
            msg (str, optional): Descriptive message for the timed operation.
                If provided, will be included in the begin/end log messages
                to help identify what is being timed. Defaults to None.
            loglevel (int, optional): Python logging level for timer output.
                Common values include:
                - logging.DEBUG (10): Detailed debugging information
                - logging.INFO (20): General information (default)
                - logging.WARNING (30): Warning messages
                - logging.ERROR (40): Error messages
                Defaults to logging.INFO.
        """
        self.loglevel = loglevel
        self.msg = msg

    def __enter__(self) -> None:
        """
        Enter the timing context and start the timer.

        This method is automatically called when entering the 'with' statement.
        It records the current timestamp, increments the nesting level for
        proper log indentation, and optionally logs a begin message.

        Returns:
            None: This context manager doesn't return a value to the 'as' clause.
                  All timing information is handled internally and logged automatically.
        """
        self.start_time = time.time()
        timeit_context.nest_level += 1
        if self.msg is not None:
            self._log("  " * timeit_context.nest_level + "[BEGIN] " + str(self.msg))

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Exit the timing context and log the elapsed time.

        This method is automatically called when exiting the 'with' statement,
        whether through normal completion or exception. It calculates the total
        elapsed time and logs the results with proper nesting indentation.

        Args:
            exc_type (type): Exception type if an exception occurred, None otherwise.
            exc_val (Exception): Exception instance if an exception occurred, None otherwise.
            exc_tb (traceback): Exception traceback if an exception occurred, None otherwise.

        Returns:
            None: Does not suppress exceptions (returns None/False implicitly).
                  Any exceptions that occurred in the timed block will propagate normally.
        """
        end_time = time.time()
        elapsed_time = end_time - self.start_time
        self._log(
            "  " * timeit_context.nest_level
            + "[END]   "
            + str(f"Elapsed time: {elapsed_time:.2f}s")
        )
        timeit_context.nest_level -= 1



from difflib import get_close_matches
from typing import Any, Iterable, List, Optional, TypeVar, Union

T = TypeVar("T")

__all__ = [
    "first",
    "has_length",
    "join_lists",
    "attr_equal",
    "validate_and_suggest_corrections",
]


def first(iterable: Iterable[T], default: Optional[T] = None) -> Optional[T]:
    """
    Return the first element of an iterable.

    Args:
        iterable: The iterable to get the first element from.
        default: The value to return if the iterable is empty. If None and
                the iterable is empty, raises StopIteration.

    Returns:
        The first element of the iterable, or the default value if empty.

    Raises:
        StopIteration: If the iterable is empty and no default is provided.
        TypeError: If the object is not iterable.
    """
    try:
        iterator = iter(iterable)
        return next(iterator)
    except StopIteration:
        if default is not None:
            return default
        raise
    except TypeError as e:
        raise TypeError(
            f"Object of type {type(iterable).__name__} is not iterable"
        ) from e


def has_length(obj: Any) -> bool:
    """
    Check if an object has a length (implements __len__) and len() works correctly.

    Args:
        obj: The object to check for length support.

    Returns:
        bool: True if the object supports len() and doesn't raise an error,
              False otherwise.
    """
    if obj is None:
        return False

    try:
        # Check if __len__ method exists
        if not hasattr(obj, "__len__"):
            return False

        # Try to get the length - this will raise TypeError for unsized objects
        length = len(obj)

        # Verify the length is a non-negative integer
        return isinstance(length, int) and length >= 0
    except (TypeError, AttributeError):
        # TypeError: len() of unsized object
        # AttributeError: if __len__ is not callable somehow
        return False
    except Exception:
        # Any other unexpected error
        return False


def join_lists(list_of_lists: Iterable[Iterable[T]]) -> List[T]:
    """
    Flatten a collection of iterables into a single list.

    Args:
        list_of_lists: An iterable containing iterables to be flattened.

    Returns:
        List[T]: A new list containing all elements from the input iterables
                in order.

    Raises:
        TypeError: If any item in list_of_lists is not iterable.

    Examples:
        >>> join_lists([[1, 2], [3, 4], [5]])
        [1, 2, 3, 4, 5]
        >>> join_lists([])
        []
        >>> join_lists([[], [1], [], [2, 3]])
        [1, 2, 3]
    """
    if not list_of_lists:
        return []

    result = []
    for i, item in enumerate(list_of_lists):
        try:
            # Check if item is iterable (but not string, which is iterable but
            # usually not what we want to flatten character by character)
            if isinstance(item, (str, bytes)):
                raise TypeError(
                    f"Item at index {i} is a string/bytes, not a list-like iterable"
                )

            # Try to extend with the item
            result.extend(item)
        except TypeError as e:
            if "not iterable" in str(e):
                raise TypeError(
                    f"Item at index {i} (type: {type(item).__name__}) is not iterable"
                ) from e
            else:
                # Re-raise our custom error or other TypeError
                raise

    return result


def attr_equal(obj, attr: str, value):
    """
    Check if the attribute of the object is equal to the given value.
    Returns False if the attribute does not exist or is not equal to the value.

    Args:
        obj: The object to check.
        attr (str): The attribute name to check.
        value: The value to compare against.

    Returns:
        bool: True if the attribute exists and is equal to the value, False otherwise.
    """
    if not hasattr(obj, attr):
        return False
    return getattr(obj, attr) == value


def validate_and_suggest_corrections(
    obj: Any, values: Iterable[Any], *, max_suggestions: int = 3, cutoff: float = 0.6
) -> Any:
    """
    Return *obj* if it is contained in *values*.
    Otherwise raise a helpful ``ValueError`` that lists the closest matches.

    Args:
        obj : Any
            The value to validate.
        values : Iterable[Any]
            The set of allowed values.
        max_suggestions : int, optional
            How many typo-hints to include at most (default 3).
        cutoff : float, optional
            Similarity threshold for suggestions (0.0-1.0, default 0.6).

    Returns:
        The original *obj* if it is valid.

    Raises:
        ValueError: With a friendly message that points out possible typos.
    """
    # Normalise to a list so we can reuse it
    value_list = list(values)

    if obj in value_list:
        return obj

    # Build suggestions
    str_values = list(map(str, value_list))
    matches = get_close_matches(str(obj), str_values, n=max_suggestions, cutoff=cutoff)

    msg = f"Invalid value {obj!r}. Allowed values: {value_list}"
    if matches:
        msg += f". Did you mean {', '.join(repr(m) for m in matches)}?"
    raise ValueError(msg)



import logging
import os
from collections import OrderedDict
from copy import deepcopy
from typing import Optional

import torch

from fusion_bench.modelpool import BaseModelPool
from fusion_bench.utils.parameters import count_parameters
from fusion_bench.utils.state_dict_arithmetic import (
    state_dict_add,
    state_dict_mul,
    state_dict_sub,
)

from .base_algorithm import SuperposedAlgorithmBase, compare_models

log = logging.getLogger(__name__)


class SuperposedTaskArithmeticAlgorithm(
    SuperposedAlgorithmBase,
):
    _config_mapping = SuperposedAlgorithmBase._config_mapping | {
        "scaling_factor": "scaling_factor",
        "model_path": "model_path",
    }

    def __init__(
        self,
        scaling_factor: float,
        model_path: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.scaling_factor = scaling_factor
        self.model_path = model_path

    @torch.no_grad()
    def run(self, modelpool: BaseModelPool):
        if not isinstance(modelpool, BaseModelPool):
            modelpool = BaseModelPool(models=modelpool)

        log.info("Compressing models using superposed task arithmetic.")
        task_vector = None
        with self.profile("load model"):
            pretrained_model = modelpool.load_model("_pretrained_")

        # Calculate the task vector superposition
        task_vectors = {}
        models = {}
        for model_name in modelpool.model_names:
            with self.profile("load model"):
                model = modelpool.load_model(model_name)
            for layer_name, layer in model.state_dict(keep_vars=True).items():
                if self.verbose >= 1:
                    log.info(f"{layer_name} | {layer.shape}")
            task_vector = state_dict_sub(
                model.state_dict(keep_vars=True),
                pretrained_model.state_dict(keep_vars=True),
            )
            task_vectors[model_name] = task_vector

        with self.profile("compress and retrieve"):
            retrieved_task_vectors, metadata = self._compress_and_retrieve(
                deepcopy(task_vectors), mode="superposed_task_arithmetic"
            )
        with self.profile("retrieve models"):
            for model_name in modelpool.model_names:
                retrieved_task_vector = state_dict_mul(
                    retrieved_task_vectors[model_name], self.scaling_factor
                )
                retrieved_state_dict = state_dict_add(
                    pretrained_model.state_dict(keep_vars=True), retrieved_task_vector
                )
                retrieved_model = deepcopy(pretrained_model)
                # FIXME: for 'all' mode
                for k, v in retrieved_state_dict.items():
                    if v.shape[0] == 1:
                        retrieved_state_dict[k] = v.squeeze(0)
                retrieved_model.load_state_dict(retrieved_state_dict)
                models[model_name] = retrieved_model

                if self.debug >= 1:
                    with self.profile("metadata"):
                        model = modelpool.load_model(model_name)
                        if torch.cuda.is_available():
                            retrieved_state_dict = {
                                k: v.cuda() for k, v in retrieved_state_dict.items()
                            }
                            retrieved_task_vectors[model_name] = {
                                k: v.cuda()
                                for k, v in retrieved_task_vectors[model_name].items()
                            }
                            task_vectors[model_name] = {
                                k: v.cuda() for k, v in task_vectors[model_name].items()
                            }
                            model_state_dict = {
                                k: v.cuda()
                                for k, v in model.state_dict(keep_vars=True).items()
                            }
                        # target_layers = metadata['target_layers']
                        metadata["task_vector_retrieval_similarity"][model_name] = (
                            compare_models(
                                retrieved_task_vectors[model_name],
                                task_vectors[model_name],
                            )
                        )
                        metadata["task_vector_svd_subspace_similarities"][
                            model_name
                        ] = self._compute_svd_subspace_similarities(
                            task_vectors[model_name], retrieved_task_vectors[model_name]
                        )
                        # overall retrieval performance
                        metadata["model_retrieval_similarity"][model_name] = (
                            compare_models(retrieved_state_dict, model_state_dict)
                        )
                        metadata["model_svd_subspace_similarities"][model_name] = (
                            self._compute_svd_subspace_similarities(
                                model_state_dict, retrieved_state_dict
                            )
                        )
                        # delete the cuda tensors
                        del (
                            retrieved_state_dict,
                            retrieved_task_vectors[model_name],
                            task_vectors[model_name],
                            model_state_dict,
                        )

        with self.profile("metadata"):
            if self.debug >= 0:
                (
                    metadata["trainable_param_count_pretrained_model"],
                    metadata["active_param_count_pretrained_model"],
                ) = count_parameters(pretrained_model)
                (
                    metadata["trainable_param_count_retrieved_model"],
                    metadata["active_param_count_retrieved_model"],
                ) = count_parameters(models[modelpool.model_names[0]])
                metadata["nonzero_parameter_count"] += metadata[
                    "active_param_count_pretrained_model"
                ]
                metadata["total_gb_retrieved"] += metadata["total_gb_original"]
                print(
                    f"Total storage (Gbs) for retrieval and original: {metadata['total_gb_retrieved']} | {metadata['total_gb_original']}"
                )

        if self.model_path is not None:
            os.makedirs(os.path.dirname(self.model_path), exist_ok=True)
            torch.save(models, self.model_path)

        self.print_profile_summary()
        return {"models": models, "metadata": metadata}


class SuperposedTaskArithmeticLoRAAlgorithm(
    SuperposedAlgorithmBase,
):
    _config_mapping = SuperposedAlgorithmBase._config_mapping | {
        "scaling_factor": "scaling_factor",
        "model_path": "model_path",
    }

    def __init__(
        self,
        scaling_factor: float,
        model_path: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.scaling_factor = scaling_factor
        self.model_path = model_path

    @torch.no_grad()
    def run(self, modelpool: BaseModelPool):
        if not isinstance(modelpool, BaseModelPool):
            modelpool = BaseModelPool(models=modelpool)

        log.info("Compressing models using superposed task arithmetic.")
        task_vector = None
        with self.profile("load model"):
            pretrained_model = modelpool.load_model("_pretrained_")

        # Calculate the task vector superposition
        loras = {}
        models = {}
        for model_name in modelpool.model_names:
            with self.profile("load model"):
                model = modelpool.load_model(model_name)
            for layer_name, layer in model.items():
                if self.verbose >= 1:
                    log.info(f"{layer_name} | {layer.shape}")
            # task_vector = state_dict_sub(
            #     model.state_dict(keep_vars=True),
            #     pretrained_model.state_dict(keep_vars=True),
            # )
            loras[model_name] = model

        with self.profile("compress and retrieve"):
            retrieved_loras, metadata = self._compress_and_retrieve(
                deepcopy(loras), mode="superposed_task_arithmetic"
            )
        with self.profile("retrieve models"):
            for model_name in modelpool.model_names:
                retrieved_lora = retrieved_loras[model_name]
                # retrieved_lora = state_dict_mul(retrieved_loras[model_name], self.config.scaling_factor)
                # retrieved_state_dict = state_dict_add(pretrained_model.state_dict(keep_vars=True), retrieved_lora)
                retrieved_model = deepcopy(pretrained_model)
                sd = retrieved_model.state_dict(keep_vars=True)
                # for layer_name, layer in sd.items():
                #     print(layer_name)
                # manually merge the lora back
                lora_weights = {}
                lora_weights_ready_to_merge = OrderedDict()
                for layer_name, layer in retrieved_lora.items():
                    parts = layer_name.split(".")
                    # print(parts)
                    base_name = ".".join(parts[2:-2] + [parts[-1]])
                    if base_name not in lora_weights:
                        lora_weights[base_name] = []
                    lora_weights[base_name].append(layer)
                for base_name, layers in lora_weights.items():
                    lora_weight = layers[-1] @ layers[0]
                    # sd[base_name] += lora_weight
                    lora_weights_ready_to_merge[base_name] = lora_weight

                retrieved_lora_ready = state_dict_mul(
                    lora_weights_ready_to_merge, self.config.scaling_factor
                )
                for layer_name, layer in retrieved_lora_ready.items():
                    sd[layer_name] += layer
                retrieved_model.load_state_dict(sd)
                models[model_name] = retrieved_model

                # # FIXME: for 'all' mode
                # for k, v in retrieved_state_dict.items():
                #     if v.shape[0] == 1:
                #         retrieved_state_dict[k] = v.squeeze(0)
                # retrieved_model.load_state_dict(sd)
                # models[model_name] = retrieved_model

                if self.debug >= 1:
                    with self.profile("metadata"):
                        model = modelpool.load_model(model_name)
                        if torch.cuda.is_available():
                            retrieved_state_dict = {
                                k: v.cuda() for k, v in retrieved_state_dict.items()
                            }
                            retrieved_loras[model_name] = {
                                k: v.cuda()
                                for k, v in retrieved_loras[model_name].items()
                            }
                            loras[model_name] = {
                                k: v.cuda() for k, v in loras[model_name].items()
                            }
                            model_state_dict = {
                                k: v.cuda()
                                for k, v in model.state_dict(keep_vars=True).items()
                            }
                        # focus on the superposition retrieval performance on the target layers
                        target_layers = metadata["target_layers"]
                        metadata["lora_retrieval_similarity"][model_name] = (
                            compare_models(
                                retrieved_loras[model_name],
                                loras[model_name],
                                target_layers,
                            )
                        )
                        metadata["lora_svd_subspace_similarities"][model_name] = (
                            self._compute_svd_subspace_similarities(
                                loras[model_name],
                                retrieved_loras[model_name],
                                target_layers,
                            )
                        )
                        # overall retrieval performance
                        metadata["model_retrieval_similarity"][model_name] = (
                            compare_models(retrieved_state_dict, model_state_dict)
                        )
                        metadata["model_svd_subspace_similarities"][model_name] = (
                            self._compute_svd_subspace_similarities(
                                model_state_dict, retrieved_state_dict
                            )
                        )
                        # delete the cuda tensors
                        del (
                            retrieved_state_dict,
                            retrieved_loras[model_name],
                            loras[model_name],
                            model_state_dict,
                        )

        with self.profile("metadata"):
            if self.debug >= 0:
                (
                    metadata["trainable_param_count_pretrained_model"],
                    metadata["active_param_count_pretrained_model"],
                ) = count_parameters(pretrained_model)
                (
                    metadata["trainable_param_count_retrieved_model"],
                    metadata["active_param_count_retrieved_model"],
                ) = count_parameters(models[modelpool.model_names[0]])
                metadata["nonzero_parameter_count"] += metadata[
                    "active_param_count_pretrained_model"
                ]
                metadata["total_gb_retrieved"] += metadata["total_gb_original"]
                print(
                    f"Total storage (Gbs) for retrieval and original: {metadata['total_gb_retrieved']} | {metadata['total_gb_original']}"
                )

        if self.model_path is not None:
            os.makedirs(os.path.dirname(self.model_path), exist_ok=True)
            torch.save(models, self.model_path)

        self.print_profile_summary()
        return {"models": models, "metadata": metadata}



"""
This module provides the `DataAdaptiveWeightEnsemblingModel` class for data-adaptive weight ensembling.

The DataAdaptiveWeightEnsemblingModel class is designed to perform data-adaptive weight ensembling
for model fusion. It supports both task-wise and layer-wise merging modes and allows for the use
of different feature extractors and processors.

Classes:
    DataAdaptiveWeightEnsemblingModel: A class for data-adaptive weight ensembling.
"""

# flake8: noqa F401
from .dawe_model import DataAdaptiveWeightEnsemblingModel



import re
from typing import Dict, List

from torch import nn


def get_param_names_to_merge(
    input_param_names: List[str], exclude_param_names_regex: list
) -> List[str]:
    """
    get the names of parameters that need to be merged

    Args:
        input_param_names: list, names of input parameters
        exclude_param_names_regex: list, regular expression of names of parameters that need to be excluded

    Returns:
        list: names of parameters that need to be merged
    """
    param_names_to_merge = []
    for param_name in input_param_names:
        exclude = any(
            [
                re.match(exclude_pattern, param_name)
                for exclude_pattern in exclude_param_names_regex
            ]
        )
        if not exclude:
            param_names_to_merge.append(param_name)
    return param_names_to_merge


def get_modules_to_merge(
    model: nn.Module, include_module_types: list
) -> Dict[str, nn.Module]:
    """
    get the model modules that need to be merged, whose type is in include_module_types

    Args:
        model: nn.Module, input model
        include_module_types: list, module types that want to include

    Returns:
        Dict[str, nn.Module]: a dictionary of modules to merge
    """
    modules_to_merge: Dict[str, nn.Module] = {}
    for module_name, module in model.named_modules():
        is_valid_type = not include_module_types or any(
            [
                isinstance(module, include_module_type)
                for include_module_type in include_module_types
            ]
        )
        if is_valid_type:
            modules_to_merge[module_name] = module
    return modules_to_merge



import fnmatch
import os
from typing import Callable, Dict, Optional, Tuple

import numpy as np
import torch
from torch.utils.data import Dataset


class NYUv2(Dataset):
    R"""
    NYUv2 dataset, 3 tasks + 1 generated useless task
    Included tasks:

        1. Semantic Segmentation,
        2. Depth prediction,
        3. Surface Normal prediction,
        4. Noise prediction [to test auxiliary learning, purely conflict gradients]

    Modified from https://github.com/lorenmt/auto-lambda/blob/main/create_dataset.py

    removed the `augmentation` arg and add `transform` args
    """

    num_out_channels = {
        "segmentation": 13,
        "depth": 1,
        "normal": 3,
        "noise": 1,
    }

    def __init__(
        self,
        root: str,
        train: bool = True,
        transform: Optional[Callable] = None,
        seg_transform: Optional[Callable] = None,
        sn_transform: Optional[Callable] = None,
        depth_transform: Optional[Callable] = None,
    ):
        """
        Initialize the NYUv2 dataset.

        Args:
            root (str): The root directory of the dataset.
            train (bool, optional): If True, use training set. If False, use validation set. Defaults to True.
            transform (Callable, optional): image transform. Defaults to None.
            seg_transform (Callable, optional): segmentation transform. Defaults to None.
            sn_transform (Callable, optional): surface normal transform. Defaults to None.
            depth_transform (Callable, optional): depth transform. Defaults to None.
        """
        self.root = os.path.expanduser(root)
        self.train = train

        self.transform = transform
        self.seg_transform = seg_transform
        self.sn_transform = sn_transform
        self.depth_transform = depth_transform

        if train:
            self.data_path = self.root + "/train"
        else:
            self.data_path = self.root + "/val"

        # calculate data length
        self.data_len = len(
            fnmatch.filter(os.listdir(self.data_path + "/image"), "*.npy")
        )
        self.noise = torch.rand(self.data_len, 1, 288, 384)

    def __getitem__(self, index) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        """
        Retrieve an item from the dataset.

        Args:
            index (int): The index of the item to retrieve.

        Returns:
            tuple: A tuple containing the image and a dictionary of task-specific outputs.
        """
        # load data from the pre-processed npy files
        image = torch.from_numpy(
            np.moveaxis(
                np.load(self.data_path + "/image/{:d}.npy".format(index)), -1, 0
            )
        ).float()
        semantic = torch.from_numpy(
            np.load(self.data_path + "/label/{:d}.npy".format(index))
        ).float()
        depth = torch.from_numpy(
            np.moveaxis(
                np.load(self.data_path + "/depth/{:d}.npy".format(index)), -1, 0
            )
        ).float()
        normal = torch.from_numpy(
            np.moveaxis(
                np.load(self.data_path + "/normal/{:d}.npy".format(index)), -1, 0
            )
        ).float()
        noise = self.noise[index].float()

        if self.transform is not None:
            image = self.transform(image)
        if self.seg_transform is not None:
            semantic = self.seg_transform(semantic)
        if self.sn_transform is not None:
            normal = self.sn_transform(normal)
        if self.depth_transform is not None:
            depth = self.depth_transform(depth)

        return image, {
            "segmentation": semantic,
            "depth": depth,
            "normal": normal,
            "noise": noise,
        }

    def __len__(self):
        return self.data_len



import os
from typing import Literal

import pandas as pd
import torch
from tqdm import tqdm

from fusion_bench import BaseAlgorithm, BaseModelPool, auto_register_config
from fusion_bench.mixins import LightningFabricMixin, SimpleProfilerMixin


class ProjectedEnergyAnalysis(
    SimpleProfilerMixin,
    LightningFabricMixin,
    BaseAlgorithm,
):
    def on_run_start(self):
        self.device = self.fabric.device

    def run(self, modelpool: BaseModelPool):
        with self.profile("model loading"):
            base_model = modelpool.load_pretrained_model()

        results = {
            "model_name": [],
            "module_index": [],
            "module_name": [],
            "projected_energy_I": [],
            "projected_energy_II": [],
            "projected_energy_II_III": [],
        }
        for model_name in tqdm(
            modelpool.model_names,
            "analyzing",
            dynamic_ncols=True,
        ):
            with self.profile("model loading"):
                finetuned_model = modelpool.load_model(model_name)

            module_index = 0
            for module_name, base_module in tqdm(
                list(base_model.named_modules()),
                "analyzing modules",
                dynamic_ncols=True,
            ):
                if isinstance(base_module, torch.nn.Linear):
                    with self.profile("weight analysis"):
                        _result = self.analyze_weight(
                            base_module.weight,
                            finetuned_model.get_submodule(module_name).weight,
                        )
                    results["model_name"].append(model_name)
                    results["module_index"].append(module_index)
                    results["module_name"].append(module_name)
                    for key, value in _result.items():
                        results[key].append(value)

                    module_index += 1

        # save results as csv
        results = pd.DataFrame(results)
        results.to_csv(
            os.path.join(self.log_dir, "projected_energy_analysis.csv"), index=True
        )

        self.print_profile_summary()
        return None

    @torch.no_grad()
    def analyze_weight(self, w: torch.Tensor, w_ft: torch.Tensor, k: int = -1):
        w = w.to(dtype=torch.float32, device=self.device)
        w_ft = w_ft.to(dtype=torch.float32, device=self.device)
        w_diff = w_ft - w

        # Perform analysis on the weight tensor
        u, s, vh = torch.linalg.svd(w, full_matrices=False)
        v = vh.T
        if k < 0:
            # find the position where the sum of singular values is larger than 50% of the total sum
            cumsum = s.cumsum(0)
            k = (cumsum < cumsum[-1] * 0.5).sum().item() + 1

        # subspace I
        w_diff_proj = self._project_subspace_low(u=u, s=s, v=v, k=k, w=w, w_ft=w_ft)
        projected_energy_I = (
            torch.linalg.norm(w_diff_proj, ord="fro") ** 2
            / torch.linalg.norm(w_diff, ord="fro") ** 2
        )

        # subspace II
        w_diff_proj = self._project_subspace_high(u=u, s=s, v=v, k=k, w=w, w_ft=w_ft)
        projected_energy_II = (
            torch.linalg.norm(w_diff_proj, ord="fro") ** 2
            / torch.linalg.norm(w_diff, ord="fro") ** 2
        )

        ## subspace II+III
        u, s, vh = torch.linalg.svd(w, full_matrices=True)
        v = vh.T
        w_diff_proj = self._project_subspace_high(u=u, s=s, v=v, k=k, w=w, w_ft=w_ft)
        projected_energy_II_III = (
            torch.linalg.norm(w_diff_proj, ord="fro") ** 2
            / torch.linalg.norm(w_diff, ord="fro") ** 2
        )

        return {
            "projected_energy_I": projected_energy_I.item(),
            "projected_energy_II": projected_energy_II.item(),
            "projected_energy_II_III": projected_energy_II_III.item(),
        }

    def _project_subspace_low(
        self,
        u: torch.Tensor,
        s: torch.Tensor,
        v: torch.Tensor,
        k: int,
        w: torch.Tensor,
        w_ft: torch.Tensor,
    ):
        u = u[:, :k]
        s = s[:k]
        v = v[:, :k]

        w_diff = w_ft - w
        w_diff_proj = torch.linalg.multi_dot((u, u.T, w_diff, v, v.T))
        return w_diff_proj

    def _project_subspace_high(
        self,
        u: torch.Tensor,
        s: torch.Tensor,
        v: torch.Tensor,
        k: int,
        w: torch.Tensor,
        w_ft: torch.Tensor,
    ):
        u = u[:, k:]
        s = s[k:]
        v = v[:, k:]

        w_diff = w_ft - w
        w_diff_proj = torch.linalg.multi_dot((u, u.T, w_diff, v, v.T))
        return w_diff_proj



from functools import partial
from typing import Callable, Optional, Tuple, Union

import torch
import torch.nn as nn
from transformers.activations import ACT2FN
from transformers.cache_utils import Cache, HybridCache, StaticCache
from transformers.generation import GenerationMixin
from transformers.modeling_flash_attention_utils import FlashAttentionKwargs
from transformers.modeling_outputs import (
    BaseModelOutputWithPast,
    CausalLMOutputWithPast,
    SequenceClassifierOutputWithPast,
    TokenClassifierOutput,
)
from transformers.modeling_rope_utils import ROPE_INIT_FUNCTIONS, dynamic_rope_update
from transformers.modeling_utils import ALL_ATTENTION_FUNCTIONS, PreTrainedModel
from transformers.models.gemma2.modeling_gemma2 import (
    _CHECKPOINT_FOR_DOC,
    GEMMA2_INPUTS_DOCSTRING,
    Gemma2RMSNorm,
    Gemma2RotaryEmbedding,
    apply_rotary_pos_emb,
    eager_attention_forward,
)
from transformers.processing_utils import Unpack
from transformers.utils import (
    add_code_sample_docstrings,
    add_start_docstrings,
    add_start_docstrings_to_model_forward,
    can_return_tuple,
    logging,
    replace_return_docstrings,
)
from transformers.utils.deprecation import deprecate_kwarg

from fusion_bench.models.smile_moe.linear_from_hf_config import SmileLinear

from .configuration_smile_gemma2 import SmileGemma2Config

logger = logging.get_logger(__name__)

_CONFIG_FOR_DOC = "SmileGemma2Config"


class SmileGemma2MLP(nn.Module):
    def __init__(self, config: SmileGemma2Config):
        super().__init__()
        self.config = config
        self.hidden_size = config.hidden_size
        self.intermediate_size = config.intermediate_size
        # * --- replace nn.Linear with SmileLinear ---
        self.gate_proj = SmileLinear(config, self.hidden_size, self.intermediate_size)
        self.up_proj = SmileLinear(config, self.hidden_size, self.intermediate_size)
        self.down_proj = SmileLinear(config, self.intermediate_size, self.hidden_size)
        # * --- end of replacement ---
        self.act_fn = ACT2FN[config.hidden_activation]

    def forward(self, x):
        down_proj = self.down_proj(self.act_fn(self.gate_proj(x)) * self.up_proj(x))
        return down_proj


class SmileGemma2Attention(nn.Module):
    """Multi-headed attention from 'Attention Is All You Need' paper"""

    def __init__(self, config: SmileGemma2Config, layer_idx: int):
        super().__init__()
        self.config = config
        self.layer_idx = layer_idx
        self.head_dim = getattr(
            config, "head_dim", config.hidden_size // config.num_attention_heads
        )
        self.num_key_value_groups = (
            config.num_attention_heads // config.num_key_value_heads
        )
        self.scaling = config.query_pre_attn_scalar**-0.5
        self.attention_dropout = self.config.attention_dropout
        self.is_causal = True

        # * --- replace nn.Linear with SmileLinear ---
        self.q_proj = SmileLinear(
            config,
            config.hidden_size,
            config.num_attention_heads * self.head_dim,
            bias=config.attention_bias,
        )
        self.k_proj = SmileLinear(
            config,
            config.hidden_size,
            config.num_key_value_heads * self.head_dim,
            bias=config.attention_bias,
        )
        self.v_proj = SmileLinear(
            config,
            config.hidden_size,
            config.num_key_value_heads * self.head_dim,
            bias=config.attention_bias,
        )
        self.o_proj = SmileLinear(
            config,
            config.num_attention_heads * self.head_dim,
            config.hidden_size,
            bias=config.attention_bias,
        )
        # * --- end of replacement ---
        self.attn_logit_softcapping = self.config.attn_logit_softcapping
        self.sliding_window = config.sliding_window if not bool(layer_idx % 2) else None

    def forward(
        self,
        hidden_states: torch.Tensor,
        position_embeddings: Tuple[torch.Tensor, torch.Tensor],
        attention_mask: Optional[torch.Tensor],
        past_key_value: Optional[Cache] = None,
        cache_position: Optional[torch.LongTensor] = None,
        **kwargs: Unpack[FlashAttentionKwargs],
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[Tuple[torch.Tensor]]]:
        input_shape = hidden_states.shape[:-1]
        hidden_shape = (*input_shape, -1, self.head_dim)

        query_states = self.q_proj(hidden_states).view(hidden_shape).transpose(1, 2)
        key_states = self.k_proj(hidden_states).view(hidden_shape).transpose(1, 2)
        value_states = self.v_proj(hidden_states).view(hidden_shape).transpose(1, 2)

        cos, sin = position_embeddings
        query_states, key_states = apply_rotary_pos_emb(
            query_states, key_states, cos, sin
        )

        if past_key_value is not None:
            # sin and cos are specific to RoPE models; cache_position needed for the static cache
            cache_kwargs = {
                "sin": sin,
                "cos": cos,
                "cache_position": cache_position,
                "sliding_window": self.sliding_window,
            }
            key_states, value_states = past_key_value.update(
                key_states, value_states, self.layer_idx, cache_kwargs
            )

            # Here we need to slice as we use a static cache by default, but FA2 does not support it
            if (
                attention_mask is not None
                and self.config._attn_implementation == "flash_attention_2"
            ):
                seq_len = attention_mask.shape[-1]
                key_states, value_states = (
                    key_states[:, :, :seq_len, :],
                    value_states[:, :, :seq_len, :],
                )

        attention_interface: Callable = eager_attention_forward
        if self.config._attn_implementation != "eager":
            if self.config._attn_implementation == "sdpa" and kwargs.get(
                "output_attentions", False
            ):
                logger.warning_once(
                    "`torch.nn.functional.scaled_dot_product_attention` does not support `output_attentions=True`. Falling back to "
                    'eager attention. This warning can be removed using the argument `attn_implementation="eager"` when loading the model.'
                )
            else:
                attention_interface = ALL_ATTENTION_FUNCTIONS[
                    self.config._attn_implementation
                ]

        attn_output, attn_weights = attention_interface(
            self,
            query_states,
            key_states,
            value_states,
            attention_mask,
            dropout=self.attention_dropout if self.training else 0.0,
            scaling=self.scaling,
            sliding_window=self.sliding_window,
            softcap=self.attn_logit_softcapping,
            **kwargs,
        )

        attn_output = attn_output.reshape(*input_shape, -1).contiguous()
        attn_output = self.o_proj(attn_output)
        return attn_output, attn_weights


class SmileGemma2DecoderLayer(nn.Module):

    def __init__(self, config: SmileGemma2Config, layer_idx: int):
        super().__init__()
        self.hidden_size = config.hidden_size
        self.config = config
        self.is_sliding = not bool(layer_idx % 2)
        self.self_attn = SmileGemma2Attention(config=config, layer_idx=layer_idx)
        self.mlp = SmileGemma2MLP(config)
        self.input_layernorm = Gemma2RMSNorm(
            config.hidden_size, eps=config.rms_norm_eps
        )
        self.post_attention_layernorm = Gemma2RMSNorm(
            config.hidden_size, eps=config.rms_norm_eps
        )

        self.pre_feedforward_layernorm = Gemma2RMSNorm(
            config.hidden_size, eps=config.rms_norm_eps
        )
        self.post_feedforward_layernorm = Gemma2RMSNorm(
            config.hidden_size, eps=config.rms_norm_eps
        )
        self.sliding_window = config.sliding_window

    def forward(
        self,
        hidden_states: torch.Tensor,
        position_embeddings: Tuple[torch.Tensor, torch.Tensor],
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_value: Optional[Cache] = None,
        output_attentions: Optional[bool] = False,
        use_cache: Optional[bool] = False,
        cache_position: Optional[torch.LongTensor] = None,
        last_cache_position: int = 0,
        **kwargs,
    ) -> Tuple[
        torch.FloatTensor, Optional[Tuple[torch.FloatTensor, torch.FloatTensor]]
    ]:
        if (
            self.is_sliding and attention_mask is not None
        ):  # efficient SDPA and no padding
            # In prefill, we may be larger than sliding window
            effective_seq_len = max(cache_position.shape[0], self.sliding_window)
            # For FA2, the mask is 2D and is of shape [bs, processed_tokens] (not [bs, max_cache_len]),
            # thus we must slice from the right (at most `effective_seq_len` elements)
            if self.config._attn_implementation == "flash_attention_2":
                attention_mask = attention_mask[:, -effective_seq_len:]
            # Otherwise, the mask is 4D of shape [bs, 1, query_len, max_cache_len] thus we must slice
            # from the left, with an offset if we are beyond the sliding window
            else:
                min_dtype = torch.finfo(attention_mask.dtype).min
                sliding_window_mask = torch.tril(
                    torch.ones_like(attention_mask, dtype=torch.bool),
                    diagonal=-self.sliding_window,
                )
                attention_mask = torch.where(
                    sliding_window_mask, min_dtype, attention_mask
                )
                # In case we are beyond the sliding window, we need to correctly offset the mask slicing
                # `last_cache_position` is equivalent to `cache_position[-1]` but without breaking dynamo
                offset = last_cache_position - effective_seq_len
                # Should only be used when beyond the sliding window (i.e. offset > 0)
                offset = max(0, offset)
                attention_mask = attention_mask[
                    :, :, :, offset : offset + effective_seq_len
                ]

        residual = hidden_states

        hidden_states = self.input_layernorm(hidden_states)

        # Self Attention
        hidden_states, self_attn_weights = self.self_attn(
            hidden_states=hidden_states,
            position_embeddings=position_embeddings,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_value=past_key_value,
            output_attentions=output_attentions,
            use_cache=use_cache,
            cache_position=cache_position,
            **kwargs,
        )
        hidden_states = self.post_attention_layernorm(hidden_states)
        hidden_states = residual + hidden_states

        residual = hidden_states
        hidden_states = self.pre_feedforward_layernorm(hidden_states)
        hidden_states = self.mlp(hidden_states)
        hidden_states = self.post_feedforward_layernorm(hidden_states)
        hidden_states = residual + hidden_states

        outputs = (hidden_states,)

        if output_attentions:
            outputs += (self_attn_weights,)

        return outputs


class SmileGemma2PreTrainedModel(PreTrainedModel):
    config_class = SmileGemma2Config
    base_model_prefix = "model"
    supports_gradient_checkpointing = True
    _no_split_modules = ["SmileGemma2DecoderLayer"]
    _skip_keys_device_placement = ["past_key_values"]
    _supports_flash_attn_2 = True
    _supports_sdpa = True
    _supports_flex_attn = True
    _supports_cache_class = True
    _supports_quantized_cache = True
    _supports_static_cache = True
    _supports_attention_backend = True

    def _init_weights(self, module):
        std = self.config.initializer_range
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=std)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=std)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()


class SmileGemma2Model(SmileGemma2PreTrainedModel):
    """
    Transformer decoder consisting of *config.num_hidden_layers* layers. Each layer is a [`Gemma2DecoderLayer`]

    Args:
        config: Gemma2Config
    """

    def __init__(self, config: SmileGemma2Config):
        super().__init__(config)
        self.padding_idx = config.pad_token_id
        self.vocab_size = config.vocab_size

        self.embed_tokens = nn.Embedding(
            config.vocab_size, config.hidden_size, self.padding_idx
        )
        self.layers = nn.ModuleList(
            [
                SmileGemma2DecoderLayer(config, layer_idx)
                for layer_idx in range(config.num_hidden_layers)
            ]
        )
        self.norm = Gemma2RMSNorm(config.hidden_size, eps=config.rms_norm_eps)
        self.rotary_emb = Gemma2RotaryEmbedding(config=config)
        self.gradient_checkpointing = False

        # Initialize weights and apply final processing
        self.post_init()

    def get_input_embeddings(self):
        return self.embed_tokens

    def set_input_embeddings(self, value):
        self.embed_tokens = value

    @can_return_tuple
    @add_start_docstrings_to_model_forward(GEMMA2_INPUTS_DOCSTRING)
    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[HybridCache] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        cache_position: Optional[torch.LongTensor] = None,
        last_cache_position: Optional[int] = None,
        **flash_attn_kwargs: Unpack[FlashAttentionKwargs],
    ) -> BaseModelOutputWithPast:
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache

        if (input_ids is None) ^ (inputs_embeds is not None):
            raise ValueError(
                "You must specify exactly one of input_ids or inputs_embeds"
            )

        if self.gradient_checkpointing and self.training and use_cache:
            logger.warning_once(
                "`use_cache=True` is incompatible with gradient checkpointing. Setting `use_cache=False`."
            )
            use_cache = False

        if inputs_embeds is None:
            inputs_embeds = self.embed_tokens(input_ids)

        if use_cache and past_key_values is None and not self.training:
            batch_size, seq_len, _ = inputs_embeds.shape
            # NOTE: ideally, `HybridCache` should be initialized outside the model with `layer_device_map`
            past_key_values = HybridCache(
                self.config,
                max_batch_size=batch_size,
                max_cache_len=seq_len,
                dtype=inputs_embeds.dtype,
                device=self.device,
            )

        if cache_position is None:
            past_seen_tokens = (
                past_key_values.get_seq_length() if past_key_values is not None else 0
            )
            cache_position = torch.arange(
                past_seen_tokens,
                past_seen_tokens + inputs_embeds.shape[1],
                device=inputs_embeds.device,
            )

        if position_ids is None:
            position_ids = cache_position.unsqueeze(0)

        # This is needed to correctly slice the mask without data-dependent slicing later on if using dynamo tracing
        # (retrieving the same value from `cache_position` later on would crash dynamo)
        if last_cache_position is None:
            last_cache_position = 0
            if attention_mask is not None:
                # In case a 4d mask is passed directly without using `generate`, we have to rely on cache_position
                # It will break dynamo tracing but there are no way around it (and it should never happen in practice)
                last_cache_position = (
                    attention_mask.shape[-1]
                    if attention_mask.dim() == 2
                    else cache_position[-1].item()
                )
        causal_mask = self._update_causal_mask(
            attention_mask,
            inputs_embeds,
            cache_position,
            past_key_values,
            output_attentions,
        )

        # embed positions
        hidden_states = inputs_embeds

        # create position embeddings to be shared across the decoder layers
        position_embeddings = self.rotary_emb(hidden_states, position_ids)

        # normalized
        # Gemma2 downcasts the below to float16, causing sqrt(3072)=55.4256 to become 55.5
        # See https://github.com/huggingface/transformers/pull/29402
        normalizer = torch.tensor(
            self.config.hidden_size**0.5, dtype=hidden_states.dtype
        )
        hidden_states = hidden_states * normalizer

        # decoder layers
        all_hidden_states = () if output_hidden_states else None
        all_self_attns = () if output_attentions else None

        for decoder_layer in self.layers[: self.config.num_hidden_layers]:
            if output_hidden_states:
                all_hidden_states += (hidden_states,)

            if self.gradient_checkpointing and self.training:
                layer_outputs = self._gradient_checkpointing_func(
                    partial(decoder_layer.__call__, **flash_attn_kwargs),
                    hidden_states,
                    position_embeddings,
                    causal_mask,
                    position_ids,
                    past_key_values,
                    output_attentions,
                    use_cache,
                    cache_position,
                    last_cache_position,
                )
            else:
                layer_outputs = decoder_layer(
                    hidden_states,
                    position_embeddings=position_embeddings,
                    attention_mask=causal_mask,
                    position_ids=position_ids,
                    past_key_value=past_key_values,
                    output_attentions=output_attentions,
                    use_cache=use_cache,
                    cache_position=cache_position,
                    last_cache_position=last_cache_position,
                    **flash_attn_kwargs,
                )

            hidden_states = layer_outputs[0]

            if output_attentions:
                all_self_attns += (layer_outputs[1],)

        hidden_states = self.norm(hidden_states)

        if output_hidden_states:
            all_hidden_states += (hidden_states,)

        return BaseModelOutputWithPast(
            last_hidden_state=hidden_states,
            past_key_values=past_key_values,
            hidden_states=all_hidden_states,
            attentions=all_self_attns,
        )

    @torch.no_grad()
    def _update_causal_mask(
        self,
        attention_mask: torch.Tensor,
        input_tensor: torch.Tensor,
        cache_position: torch.Tensor,
        past_key_values: HybridCache,
        output_attentions: bool,
    ):
        # Flash Attention currently doesn't support static cache but Gemma2 work only with static cache.
        # So we will pass in attention mask as is in any case, not only when ther's padding. Then we'll use its shape
        # to cut out keys/values trailing 0 used in static cache. This workaround should be compile compatible
        # as it doesn't cause dynamic control issues.
        if self.config._attn_implementation == "flash_attention_2":
            return attention_mask

        dtype, device = input_tensor.dtype, input_tensor.device
        sequence_length = input_tensor.shape[1]
        if isinstance(past_key_values, (HybridCache, StaticCache)):
            target_length = past_key_values.get_max_cache_shape()
        else:
            target_length = (
                attention_mask.shape[-1]
                if attention_mask is not None
                else input_tensor.shape[1]
            )

        # In case the provided `attention` mask is 2D, we generate a causal mask here (4D).
        causal_mask = self._prepare_4d_causal_attention_mask_with_cache_position(
            attention_mask,
            sequence_length=sequence_length,
            target_length=target_length,
            dtype=dtype,
            device=device,
            cache_position=cache_position,
            batch_size=input_tensor.shape[0],
        )
        return causal_mask

    @staticmethod
    def _prepare_4d_causal_attention_mask_with_cache_position(
        attention_mask: torch.Tensor,
        sequence_length: int,
        target_length: int,
        dtype: torch.dtype,
        device: torch.device,
        cache_position: torch.Tensor,
        batch_size: int,
        **kwargs,
    ):
        """
        Creates a causal 4D mask of shape `(batch_size, 1, query_length, key_value_length)` from a 2D mask of shape
        `(batch_size, key_value_length)`, or if the input `attention_mask` is already 4D, do nothing.

        Args:
            attention_mask (`torch.Tensor`):
                A 2D attention mask of shape `(batch_size, key_value_length)` or a 4D attention mask of shape
                `(batch_size, 1, query_length, key_value_length)`.
            sequence_length (`int`):
                The sequence length being processed.
            target_length (`int`):
                The target length: when generating with static cache, the mask should be as long as the static cache,
                to account for the 0 padding, the part of the cache that is not filled yet.
            dtype (`torch.dtype`):
                The dtype to use for the 4D attention mask.
            device (`torch.device`):
                The device to place the 4D attention mask on.
            cache_position (`torch.Tensor`):
                Indices depicting the position of the input sequence tokens in the sequence.
            batch_size (`torch.Tensor`):
                Batch size.
        """
        if attention_mask is not None and attention_mask.dim() == 4:
            # In this case we assume that the mask comes already in inverted form and requires no inversion or slicing.
            causal_mask = attention_mask
        else:
            min_dtype = torch.finfo(dtype).min
            causal_mask = torch.full(
                (sequence_length, target_length),
                fill_value=min_dtype,
                dtype=dtype,
                device=device,
            )
            if sequence_length != 1:
                causal_mask = torch.triu(causal_mask, diagonal=1)
            causal_mask *= torch.arange(
                target_length, device=device
            ) > cache_position.reshape(-1, 1)
            causal_mask = causal_mask[None, None, :, :].expand(batch_size, 1, -1, -1)
            if attention_mask is not None:
                causal_mask = (
                    causal_mask.clone()
                )  # copy to contiguous memory for in-place edit
                mask_length = attention_mask.shape[-1]
                padding_mask = causal_mask[:, :, :, :mask_length] + attention_mask[
                    :, None, None, :
                ].to(causal_mask.device)
                padding_mask = padding_mask == 0
                causal_mask[:, :, :, :mask_length] = causal_mask[
                    :, :, :, :mask_length
                ].masked_fill(padding_mask, min_dtype)

        return causal_mask


class SmileGemma2ForCausalLM(SmileGemma2PreTrainedModel, GenerationMixin):
    _tied_weights_keys = ["lm_head.weight"]
    _tp_plan = {"lm_head": "colwise_rep"}
    _pp_plan = {"lm_head": (["hidden_states"], ["logits"])}

    def __init__(self, config):
        super().__init__(config)
        self.model = SmileGemma2Model(config)
        self.vocab_size = config.vocab_size
        self.lm_head = nn.Linear(config.hidden_size, config.vocab_size, bias=False)

        # Initialize weights and apply final processing
        self.post_init()

    def get_input_embeddings(self):
        return self.model.embed_tokens

    def set_input_embeddings(self, value):
        self.model.embed_tokens = value

    def get_output_embeddings(self):
        return self.lm_head

    def set_output_embeddings(self, new_embeddings):
        self.lm_head = new_embeddings

    def set_decoder(self, decoder):
        self.model = decoder

    def get_decoder(self):
        return self.model

    @can_return_tuple
    @deprecate_kwarg("num_logits_to_keep", version="4.50", new_name="logits_to_keep")
    @add_start_docstrings_to_model_forward(GEMMA2_INPUTS_DOCSTRING)
    @replace_return_docstrings(
        output_type=CausalLMOutputWithPast, config_class=_CONFIG_FOR_DOC
    )
    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[HybridCache] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        cache_position: Optional[torch.LongTensor] = None,
        logits_to_keep: Union[int, torch.Tensor] = 0,
        **loss_kwargs,
    ) -> CausalLMOutputWithPast:
        r"""
            labels (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*):
                Labels for computing the masked language modeling loss. Indices should either be in `[0, ...,
                config.vocab_size]` or -100 (see `input_ids` docstring). Tokens with indices set to `-100` are ignored
                (masked), the loss is only computed for the tokens with labels in `[0, ..., config.vocab_size]`.

            logits_to_keep (`int` or `torch.Tensor`, *optional*):
                If an `int`, compute logits for the last `logits_to_keep` tokens. If `0`, calculate logits for all
                `input_ids` (special case). Only last token logits are needed for generation, and calculating them only for that
                token can save memory, which becomes pretty significant for long sequences or large vocabulary size.
                If a `torch.Tensor`, must be 1D corresponding to the indices to keep in the sequence length dimension.
                This is useful when using packed tensor format (single dimension for batch and sequence length).

        Returns:

        Example:

        ```python
        >>> from transformers import AutoTokenizer, Gemma2ForCausalLM

        >>> model = Gemma2ForCausalLM.from_pretrained("google/gemma-2-9b")
        >>> tokenizer = AutoTokenizer.from_pretrained("google/gemma-2-9b")

        >>> prompt = "What is your favorite condiment?"
        >>> inputs = tokenizer(prompt, return_tensors="pt")

        >>> # Generate
        >>> generate_ids = model.generate(inputs.input_ids, max_length=30)
        >>> tokenizer.batch_decode(generate_ids, skip_special_tokens=True, clean_up_tokenization_spaces=False)[0]
        "What is your favorite condiment?"
        ```"""

        if self.training and self.config._attn_implementation != "eager":
            logger.warning_once(
                "It is strongly recommended to train Gemma2 models with the `eager` attention implementation "
                f"instead of `{self.config._attn_implementation}`. Use `eager` with `AutoModelForCausalLM.from_pretrained('<path-to-checkpoint>', attn_implementation='eager')`."
            )
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        # decoder outputs consists of (dec_features, layer_state, dec_hidden, dec_attn)
        outputs: BaseModelOutputWithPast = self.model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            cache_position=cache_position,
            **loss_kwargs,
        )

        hidden_states = outputs.last_hidden_state
        # Only compute necessary logits, and do not upcast them to float if we are not computing the loss
        slice_indices = (
            slice(-logits_to_keep, None)
            if isinstance(logits_to_keep, int)
            else logits_to_keep
        )
        logits = self.lm_head(hidden_states[:, slice_indices, :])
        if self.config.final_logit_softcapping is not None:
            logits = logits / self.config.final_logit_softcapping
            logits = torch.tanh(logits)
            logits = logits * self.config.final_logit_softcapping

        loss = None
        if labels is not None:
            loss = self.loss_function(logits, labels, self.vocab_size, **loss_kwargs)

        return CausalLMOutputWithPast(
            loss=loss,
            logits=logits,
            past_key_values=outputs.past_key_values,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )

    def prepare_inputs_for_generation(
        self,
        input_ids,
        past_key_values=None,
        attention_mask=None,
        inputs_embeds=None,
        cache_position=None,
        position_ids=None,
        use_cache=True,
        logits_to_keep=None,
        **kwargs,
    ):
        # Overwritten: has a special cache type, `HybridCache`

        model_inputs = super().prepare_inputs_for_generation(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            inputs_embeds=inputs_embeds,
            cache_position=cache_position,
            position_ids=position_ids,
            use_cache=use_cache,
            logits_to_keep=logits_to_keep,
            **kwargs,
        )

        # This is needed to correctly slice the mask without data-dependent slicing later on if using dynamo tracing
        # (retrieving the same value from `cache_position` later on would crash dynamo)
        model_inputs["last_cache_position"] = (
            attention_mask.shape[-1] if attention_mask is not None else 0
        )
        if logits_to_keep is None:
            _ = model_inputs.pop("logits_to_keep", None)

        if (
            isinstance(past_key_values, HybridCache)
            and attention_mask.ndim == 2
            and not self.config._attn_implementation == "flash_attention_2"
        ):
            if model_inputs["inputs_embeds"] is not None:
                batch_size, sequence_length, _ = model_inputs["inputs_embeds"].shape
                device = model_inputs["inputs_embeds"].device
            else:
                batch_size, sequence_length = model_inputs["input_ids"].shape
                device = model_inputs["input_ids"].device

            attention_mask = (
                self.model._prepare_4d_causal_attention_mask_with_cache_position(
                    attention_mask,
                    sequence_length=sequence_length,
                    target_length=past_key_values.get_max_cache_shape(),
                    dtype=self.lm_head.weight.dtype,
                    device=device,
                    cache_position=cache_position,
                    batch_size=batch_size,
                )
            )
            model_inputs["attention_mask"] = attention_mask

        return model_inputs


class SmileGemma2ForSequenceClassification(SmileGemma2PreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels
        self.model = SmileGemma2Model(config)
        self.score = nn.Linear(config.hidden_size, self.num_labels, bias=False)

        # Initialize weights and apply final processing
        self.post_init()

    def get_input_embeddings(self):
        return self.model.embed_tokens

    def set_input_embeddings(self, value):
        self.model.embed_tokens = value

    @can_return_tuple
    @add_start_docstrings_to_model_forward(GEMMA2_INPUTS_DOCSTRING)
    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Cache] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
    ) -> SequenceClassifierOutputWithPast:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size,)`, *optional*):
            Labels for computing the sequence classification/regression loss. Indices should be in `[0, ...,
            config.num_labels - 1]`. If `config.num_labels == 1` a regression loss is computed (Mean-Square loss), If
            `config.num_labels > 1` a classification loss is computed (Cross-Entropy).
        """

        transformer_outputs: BaseModelOutputWithPast = self.model(
            input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )
        hidden_states = transformer_outputs.last_hidden_state
        logits = self.score(hidden_states)

        if input_ids is not None:
            batch_size = input_ids.shape[0]
        else:
            batch_size = inputs_embeds.shape[0]

        if self.config.pad_token_id is None and batch_size != 1:
            raise ValueError(
                "Cannot handle batch sizes > 1 if no padding token is defined."
            )
        if self.config.pad_token_id is None:
            last_non_pad_token = -1
        elif input_ids is not None:
            # To handle both left- and right- padding, we take the rightmost token that is not equal to pad_token_id
            non_pad_mask = (input_ids != self.config.pad_token_id).to(
                logits.device, torch.int32
            )
            token_indices = torch.arange(
                input_ids.shape[-1], device=logits.device, dtype=torch.int32
            )
            last_non_pad_token = (token_indices * non_pad_mask).argmax(-1)
        else:
            last_non_pad_token = -1
            logger.warning_once(
                f"{self.__class__.__name__} will not detect padding tokens in `inputs_embeds`. Results may be "
                "unexpected if using padding tokens in conjunction with `inputs_embeds.`"
            )

        pooled_logits = logits[
            torch.arange(batch_size, device=logits.device), last_non_pad_token
        ]

        loss = None
        if labels is not None:
            loss = self.loss_function(
                logits=logits,
                labels=labels,
                pooled_logits=pooled_logits,
                config=self.config,
            )

        return SequenceClassifierOutputWithPast(
            loss=loss,
            logits=pooled_logits,
            past_key_values=transformer_outputs.past_key_values,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
        )


class SmileGemma2ForTokenClassification(SmileGemma2PreTrainedModel):
    def __init__(self, config: SmileGemma2Config):
        super().__init__(config)
        self.num_labels = config.num_labels
        self.model = SmileGemma2Model(config)
        if getattr(config, "classifier_dropout", None) is not None:
            classifier_dropout = config.classifier_dropout
        elif getattr(config, "hidden_dropout", None) is not None:
            classifier_dropout = config.hidden_dropout
        else:
            classifier_dropout = 0.1
        self.dropout = nn.Dropout(classifier_dropout)
        self.score = nn.Linear(config.hidden_size, config.num_labels)

        # Initialize weights and apply final processing
        self.post_init()

    def get_input_embeddings(self):
        return self.model.embed_tokens

    def set_input_embeddings(self, value):
        self.model.embed_tokens = value

    @can_return_tuple
    @add_start_docstrings_to_model_forward(GEMMA2_INPUTS_DOCSTRING)
    @add_code_sample_docstrings(
        checkpoint=_CHECKPOINT_FOR_DOC,
        output_type=TokenClassifierOutput,
        config_class=_CONFIG_FOR_DOC,
    )
    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Cache] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
    ) -> TokenClassifierOutput:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size,)`, *optional*):
            Labels for computing the sequence classification/regression loss. Indices should be in `[0, ...,
            config.num_labels - 1]`. If `config.num_labels == 1` a regression loss is computed (Mean-Square loss), If
            `config.num_labels > 1` a classification loss is computed (Cross-Entropy).
        """

        outputs: BaseModelOutputWithPast = self.model(
            input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )
        sequence_output = outputs.last_hidden_state
        sequence_output = self.dropout(sequence_output)
        logits = self.score(sequence_output)

        loss = None
        if labels is not None:
            loss = self.loss_function(logits, labels, self.config)

        return TokenClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )


__all__ = [
    "SmileGemma2PreTrainedModel",
    "SmileGemma2Model",
    "SmileGemma2ForCausalLM",
    "SmileGemma2ForSequenceClassification",
    "SmileGemma2ForTokenClassification",
]



import time
from typing import Optional

import lightning as L

from fusion_bench.utils.pylogger import get_rankzero_logger

log = get_rankzero_logger(__name__)


def seed_everything_by_time(fabric: Optional[L.Fabric] = None):
    """
    Set seed for all processes by time.
    """
    # set seed for all processes
    if fabric is None or fabric.is_global_zero:
        seed = int(time.time())
    else:
        seed = None
    if fabric is not None:
        log.debug(f"Broadcasting seed `{seed}` to all processes")
        fabric.barrier()
        seed = fabric.broadcast(seed, src=0)
    L.seed_everything(seed)



import functools
import logging
from copy import deepcopy
from typing import (  # noqa: F401
    Any,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    TypeVar,
)

import torch
from torch import Tensor, nn
from torch.func import functional_call

from fusion_bench.models.utils import del_attr, get_attr, set_attr
from fusion_bench.utils.type import StateDictType, TorchModelType

__all__ = ["get_layer_wise_weights", "fuse_weights", "LayerWiseMergedModel"]

log = logging.getLogger(__name__)


def get_layer_wise_weights(
    num_models: int,
    num_layers: int,
    init_values: float = None,
    dtype: torch.dtype = torch.float32,
):
    """
    Return a tensor of layer-wise weights for the given number of models and layers.

    Args:
        num_models (int): The number of models to fuse.
        num_layers (int): The number of layers in each model.
        init_values (float, optional): The initial value for each weight. Defaults to 1.0 / num_models.
        dtype (torch.dtype): dtype of weights. This should be the same with model dtype.

    Returns:
        Tensor: A tensor of shape (num_models, num_layers) containing the layer-wise weights.
    """
    assert num_models >= 1, f"num_models must be >= 1, got {num_models}"
    assert num_layers >= 1, f"num_layers must be >= 1, got {num_layers}"
    if init_values is None:
        init_values = 1.0 / num_models
    return torch.full((num_models, num_layers), init_values, dtype=dtype)


def _fuse_weights(layer_wise_weight: Tensor, tensors: List[Tensor]) -> Tensor:
    """
    Fuse the layer-wise weights with the given state dictionaries.

    Args:
        layer_wise_weight (Tensor): A tensor of shape (num_models,) containing the layer-wise weights.
        state_dicts (List[Tensor]): A list of state dictionaries, each containing the weights for a single layer.

    Returns:
        Tensor: A tensor of shape (num_params,) containing the fused weights.
    """
    assert len(layer_wise_weight) == len(
        tensors
    ), f"layer_wise_weight.shape={layer_wise_weight.shape}, len(tensors)={len(tensors)}"
    return sum(
        layer_wise_weight[i] * w.to(layer_wise_weight.device)
        for i, w in enumerate(tensors)
    )


def fuse_weights(
    layer_wise_weight: Tensor, state_dicts: List[StateDictType]
) -> StateDictType:
    """
    Fuse the weights of multiple models using layer-wise fusion.

    Args:
        layer_wise_weight (Tensor): A tensor of shape (num_models, num_layers) representing the weight of each layer for each model.
        state_dicts (List[StateDict]): A list of state dictionaries, one for each model.

    Returns:
        A dictionary mapping each weight tensor key to the fused weight tensor.
    """
    num_models = len(state_dicts)
    num_layers = len(state_dicts[0])
    assert layer_wise_weight.shape == (
        num_models,
        num_layers,
    ), f"layer_wise_weight.shape={layer_wise_weight.shape}, expected (num_models, num_layers): ({num_models}, {num_layers})"
    return {
        k: _fuse_weights(
            layer_wise_weight[:, i], [state_dict[k] for state_dict in state_dicts]
        )
        for i, k in enumerate(state_dicts[0].keys())
    }


class LayerWiseMergedModel(nn.Module, Generic[TorchModelType]):
    _merged_state_dict: StateDictType = None

    def __init__(
        self,
        layer_wise_weight: Tensor,
        pretrained_model: TorchModelType,
        finetuned_models: List[TorchModelType],
        clamp_weights: bool = True,
        tie_weights: bool = False,
        strict: bool = True,
        sparsity_ratio: Optional[float] = None,
        normalized_merging_weights: bool = False,
    ):
        R"""
        This class wraps a pretrained model and a list of finetuned models, and merges the weights of the finetuned models into the pretrained model using layer-wise fusion.

        Reference:

            (ICLR 2024) Yang E, Wang Z, Shen L, et al. Adamerging: Adaptive model merging for multi-task learning. https://arxiv.org/pdf/2310.02575

        Args:
            layer_wise_weight (Tensor): A tensor of shape (num_models, num_layers) representing the weight of each layer for each model.
            pretrained_model (nn.Module): The pretrained model to merge the weights into.
            finetuned_models (List[nn.Module]): A list of finetuned models to merge the weights from. This should have the same architecture as the pretrained model. We use these models to compute the task vectors.
            clamp_weights (bool, optional): If True, the layer-wise weights will be clamped to [0, 1]. Defaults to True.
            tie_weights (bool, optional): This option passes the `tie_weights` argument to the `functional_call` function. Defaults to False.
            strict (bool, optional): This option passes the `strict` argument to the `functional_call` function. Defaults to True.
            sparsity_ratio (float, optional): If `sparsity_ratio` is provided, the task vector will be pruned before merging. A high spasity level can save the memory usage during merging.
            normalized_merging_weights (bool, optional): If True, the layer-wise weights will be normalized for each layer, so that the sum of weights across models for each layer is 1. Defaults to False.
        """
        super().__init__()
        self.clamp_weights = clamp_weights
        self.tie_weights = tie_weights
        self.strict = strict
        self.sparsity_ratio = sparsity_ratio
        self.nromalized_merging_weights = normalized_merging_weights

        self.merge_weight = nn.Parameter(layer_wise_weight, requires_grad=True)

        for name, param in pretrained_model.named_parameters():
            if not param.requires_grad:
                for m in finetuned_models:
                    del_attr(m, name.split("."))
            else:
                for m in finetuned_models:
                    get_attr(m, name.split(".")).data = (
                        get_attr(m, name.split(".")) - param
                    )

        self.pretrained_model = pretrained_model.requires_grad_(False)
        for m in finetuned_models:
            m.requires_grad_(False)

        self.task_vectors = nn.ModuleList(finetuned_models)

        # if `sparisty_ratio` is given, pruning the task vectors.
        if sparsity_ratio is not None:
            from fusion_bench.method.pruning.prune_utils import (
                unstructured_magnitude_prune_,
            )

            for name, param in self.task_vectors.named_parameters():
                if param.dim() != 2:
                    continue
                print(f"pruning {name}")
                pruned_param = unstructured_magnitude_prune_(
                    param.data.clone(), torch.abs, sparsity_ratio=sparsity_ratio
                )
                set_attr(
                    self.task_vectors,
                    name.split("."),
                    nn.Parameter(pruned_param.to_sparse(), requires_grad=False),
                )

    @property
    def forward_model(self):
        return functools.partial(
            functional_call,
            self.pretrained_model,
            self._merged_state_dict,
            tie_weights=self.tie_weights,
            strict=self.strict,
        )

    def merge_and_unload(self, task_vector_mask: Optional[Dict[str, Tensor]] = None):
        self.merge_weights(task_vector_mask=task_vector_mask)
        self.pretrained_model.load_state_dict(self._merged_state_dict)
        return self.pretrained_model

    def merge_weights(self, task_vector_mask: Optional[Dict[str, Tensor]] = None):
        """
        Merges the weights of the model.
        Call this after each update step.
        """
        if self.clamp_weights:
            layer_wise_weight = self.merge_weight.clamp(0, 1)
        else:
            layer_wise_weight = self.merge_weight
        if self.nromalized_merging_weights:
            # normalize the weights for each layer, so that the sum of weights across models for each layer is 1.
            layer_wise_weight = layer_wise_weight.softmax(dim=0)

        state_dict = self.pretrained_model.state_dict(keep_vars=True)
        # shape of layer_wise_weight: (num_models, num_layers)
        for weight, task_vector in zip(layer_wise_weight, self.task_vectors):
            assert len(list(task_vector.named_parameters())) == weight.size(0)
            if task_vector_mask is not None:
                weight = [
                    w * task_vector_mask[name]
                    for w, (name, param) in zip(weight, task_vector.named_parameters())
                ]
            for w, (name, param) in zip(weight, task_vector.named_parameters()):
                state_dict[name] = state_dict[name] + param * w
        self._merged_state_dict = state_dict

        return state_dict

    def forward(self, *args, **kwargs):
        if self._merged_state_dict is None:
            self.merge_weights()
        return self.forward_model(args=args, kwargs=kwargs)

    # def __getattr__(self, name: str) -> Any:
    #     try:
    #         return super().__getattr__(name)
    #     except AttributeError:
    #         attr = getattr(self.model, name)
    #         if isinstance(attr, Callable):
    #             warnings.warn(
    #                 f"forwarding `{name}` to the underlying model", UserWarning
    #             )
    #         return attr

    # def __setattr__(self, name: str, value: Any) -> None:
    #     try:
    #         super().__setattr__(name, value)
    #     except AttributeError:
    #         setattr(self.model, name, value)


def merge_weights(module: nn.Module):
    """
    Merges the weights for all `LayerWiseMergedModel` instances within the given module.

    Args:
        module (nn.Module): The module to process.
    """
    if isinstance(module, LayerWiseMergedModel):
        module.merge_weights()
        return
    else:
        for submodule in module.children():
            merge_weights(submodule)


def merge_and_unload(module: nn.Module):
    """
    Merges and unloads all `LayerWiseMergedModel` instances within the given module.

    Args:
        module (nn.Module): The module to process.

    Returns:
        nn.Module: The updated module with merged weights.
    """
    if isinstance(module, LayerWiseMergedModel):
        return module.merge_and_unload()
    else:
        for name, submodule in module.named_children():
            need_merge = isinstance(submodule, LayerWiseMergedModel)
            submodule = merge_and_unload(submodule)
            if need_merge:
                setattr(module, name, submodule)
        return module


def fix_other_parts(module: nn.Module):
    """
    Sets all parameters in the module to not require gradients, except for the merge weights
    in `LayerWiseMergedModel` instances.

    Args:
        module (nn.Module): The module to process.

    Returns:
        nn.Module: The module with updated parameter requirements.
    """
    module.requires_grad_(False)
    for submodule in module.modules():
        if isinstance(submodule, LayerWiseMergedModel):
            submodule.merge_weight.requires_grad_(True)
    return module



# flake8: noqa F401
import sys
from typing import TYPE_CHECKING

from fusion_bench.utils.lazy_imports import LazyImporter

_import_structure = {
    "clip_finetune": ["ImageClassificationFineTuningForCLIP"],
    "continual_clip_finetune": ["ContinualImageClassificationFineTuningForCLIP"],
    "image_classification_finetune": [
        "ImageClassificationFineTuning",
        "ImageClassificationFineTuning_Test",
    ],
}

if TYPE_CHECKING:
    from .clip_finetune import ImageClassificationFineTuningForCLIP
    from .continual_clip_finetune import ContinualImageClassificationFineTuningForCLIP
    from .image_classification_finetune import (
        ImageClassificationFineTuning,
        ImageClassificationFineTuning_Test,
    )
else:
    sys.modules[__name__] = LazyImporter(
        __name__,
        globals()["__file__"],
        _import_structure,
    )



"""
Implementation of the Layer-Wise AdaMerging+Surgery Algorithm.

For more details, please refer to:

- (ICLR 2024) Yang, et.al. AdaMerging: Adaptive Model Merging for Multi-Task Learning. http://arxiv.org/abs/2310.02575
- (ICML 2024) Yang, et.al. Representation Surgery for Multi-Task Model Merging. https://arxiv.org/abs/2402.02705

Basic Example:

```shell
fusion_bench \
    method=surgery/adamerging_surgery \
    modelpool=CLIPVisionModelPool/clip-vit-base-patch32_TA8 \
    taskpool=CLIPVisionModelTaskPool/clip-vit-classification_TA8
```
"""

import copy
import functools
import gc
import logging
from typing import TYPE_CHECKING, cast

import torch
import torch.nn.functional as F
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import CLIPVisionModel

from fusion_bench.dataset.clip_dataset import CLIPDataset
from fusion_bench.method.adamerging.layer_wise_adamerging import (
    LayerWiseAdaMergingAlgorithm,
)
from fusion_bench.method.adamerging.utils import get_memory_usage
from fusion_bench.mixins import CLIPClassificationMixin
from fusion_bench.modelpool import CLIPVisionModelPool
from fusion_bench.models.surgery.surgerymodelwrapper import SurgeryModelWrapper
from fusion_bench.models.wrappers.layer_wise_fusion import LayerWiseMergedModel

log = logging.getLogger(__name__)


class CLIPLayerWiseAdaMergingSurgeryAlgorithm(
    CLIPClassificationMixin,
    LayerWiseAdaMergingAlgorithm,
):

    def on_test_time_adaptation_start(self):
        """
        Here we load the CLIP processor and construct the zero-shot classification head for each task.
        """
        self.setup_zero_shot_classification_head()

    @functools.cache
    def get_shuffled_test_loader_iter(self, task: str):
        return super().get_shuffled_test_loader_iter(
            task,
            batch_size=self.config.batch_size,
            num_workers=self.config.num_workers,
        )

    def run(self, modelpool: CLIPVisionModelPool, **kwargs):
        """
        Run the Layer-Wise AdaMerging+Surgery Algorithm.

        This method constructs the wrapped model and performs test-time adaptation if necessary. Then, it will perform surgery.

        Args:
            modelpool (ModelPool): The model pool containing the pretrained and fine-tuned models.

        Returns:
            LayerWiseMergedModel: The merged model after test-time adaptation.
        """
        log.info("Fusing models using layer-wise adaptive merging.")
        self.modelpool = modelpool
        self.log_hyperparams(self.config)

        # === Start of the AdaMerging Algorithm ===
        with self.profile("construct the wrapped model"):
            module = cast(
                LayerWiseMergedModel[CLIPVisionModel],
                self.construct_layer_wise_merged_model(modelpool),
            )

        if self.config.weights is not None:
            # skip the test-time adaptation
            merge_weight: torch.Tensor = torch.load(self.config.weights)
            module.merge_weight.data = merge_weight.to(
                device=module.merge_weight.device
            )
            merged_model = copy.deepcopy(module.merge_and_unload())
            # setup the zero-shot classification head
            self.on_test_time_adaptation_start()

        else:
            with self.profile("test-time adaptation"):
                module = self.test_time_adaptation(module)
            if self.config.get("save_merging_weights", False):
                self.save_merging_weights(
                    self.config.save_merging_weights, module.merge_weight
                )
            merged_model = copy.deepcopy(module.merge_and_unload())

        # free memory
        del module
        gc.collect()
        torch.cuda.empty_cache()

        # === Start of the Surgery Algorithm ===
        log.info("start performing Surgery")
        alpha_model = SurgeryModelWrapper(
            merged_model,
            modelpool.model_names,
            projection_dim=merged_model.config.projection_dim,
        )
        alpha_model = self.fabric.setup(alpha_model)
        log.info(get_memory_usage("after freeing memory, the memory usage of GPU is:"))

        optimizer = torch.optim.Adam(
            alpha_model.collect_trainable_params(),
            lr=1e-3,
            betas=(0.9, 0.999),
            weight_decay=0.0,
        )

        finetuned_models = {
            model_name: modelpool.load_model(model_name)
            for model_name in modelpool.model_names
        }
        for name, model in finetuned_models.items():
            model.requires_grad_(False)
            model = self.fabric.to_device(model)
            model.eval()

        for iteration in tqdm(
            range(self.config.surgery_steps),
            "surgery",
            dynamic_ncols=True,
        ):
            for dataset_name in modelpool.model_names:
                batch = next(self.get_shuffled_test_loader_iter(dataset_name))
                finetuned_feature = self.compute_features(
                    finetuned_models[dataset_name], batch[0]
                )
                features, _, _ = alpha_model.compute_surgery_features(
                    lambda model: self.compute_features(model, batch[0]),
                    dataset_name,
                )

                loss = F.l1_loss(features, finetuned_feature)

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

            if ((iteration + 1) % self.config.eval_iterations) == 0:
                # print(list(alpha_model.collect_trainable_params()))
                # Evaluate try to use the test module in fusion bench
                log.info(f"iteration: {iteration+1}")
                self._program.evaluate_merged_model(self._program.taskpool, alpha_model)

        log.info("test the result of Adamerging")
        return {"adamerging": merged_model, "surgery": alpha_model}



import functools
import json
import logging
import os
from copy import deepcopy
from functools import cached_property
from typing import Callable, List, cast

import lightning as L
import torch
from omegaconf import DictConfig, open_dict
from torch.utils.data import DataLoader
from tqdm.autonotebook import tqdm
from transformers import CLIPModel, CLIPProcessor, CLIPVisionModel

from fusion_bench.dataset import CLIPDataset, load_dataset_from_config
from fusion_bench.models.hf_clip import HFCLIPClassifier
from fusion_bench.tasks.classification import ClassificationTask
from fusion_bench.tasks.clip_classification import get_classnames_and_templates
from fusion_bench.utils.parameters import count_parameters

from .base_pool import TaskPool

os.environ["TOKENIZERS_PARALLELISM"] = "false"

log = logging.getLogger(__name__)


@functools.cache
def load_dataset_from_config_cached(dataset_config: DictConfig):
    return load_dataset_from_config(dataset_config)


class CLIPImageClassificationTask(ClassificationTask):
    """
    This class is used to define the image classification task for CLIP models.
    """

    _fabric: L.Fabric = None
    _clip_processor: CLIPProcessor = None
    #
    _taskpool: "CLIPImageClassificationTaskPool" = None

    classnames: List[str] = []
    templates: List[Callable[[str], str]] = []

    def __init__(self, task_config: DictConfig):
        self.config = task_config

        self.classnames, self.templates = get_classnames_and_templates(
            self.config["dataset"].name
        )

    @cached_property
    def test_dataset(self):
        """
        Load the test dataset for the task.
        This method is cached, so the dataset is loaded only once.
        """
        dataset_config = self.config["dataset"]
        dataset_config = self._taskpool.prepare_dataset_config(dataset_config)
        log.info(f"Loading test dataset: {dataset_config.name}")
        dataset = load_dataset_from_config_cached(dataset_config)
        dataset = CLIPDataset(dataset, self._clip_processor)
        return dataset

    @property
    def num_classes(self):
        return len(self.classnames)

    @property
    def test_loader(self):
        loader = DataLoader(
            self.test_dataset,
            batch_size=self.config["batch_size"],
            num_workers=self.config["num_workers"],
            shuffle=False,
        )
        if self._fabric is not None:
            loader = self._fabric.setup_dataloaders(loader)
        return loader

    def evaluate(self, clip_model: CLIPModel):
        """
        Evaluate the model on the image classification task.

        Args:
            clip_model (CLIPModel): The CLIP model to evaluate.

        Returns:
            dict: A dictionary containing the evaluation results.
        """
        classifier = HFCLIPClassifier(
            clip_model=clip_model, processor=self._clip_processor
        )
        classifier.set_classification_task(self.classnames, self.templates)
        if self._fabric is not None:
            classifier = self._fabric.setup_module(deepcopy(classifier))
        results = super().evaluate(classifier)
        log.info(f"Results for task {self.config.name}: {results}")
        return results


class CLIPImageClassificationTaskPool(TaskPool):
    _fabric: L.Fabric = None

    # CLIP forward model and processor
    _clip_model: CLIPModel = None
    _clip_processor: CLIPProcessor = None

    def __init__(self, taskpool_config: DictConfig):
        super().__init__(taskpool_config)

    def prepare_dataset_config(self, dataset_config: DictConfig):
        if not hasattr(dataset_config, "type"):
            with open_dict(dataset_config):
                dataset_config["type"] = self.config.dataset_type
        return dataset_config

    def prepare_task_config(self, task_config: DictConfig):
        # set default values for keys that are not present in per task configuration
        for key in ["num_workers", "batch_size", "fast_dev_run"]:
            if not hasattr(task_config, key):
                with open_dict(task_config):
                    task_config[key] = self.config[key]
        return task_config

    @property
    def clip_model(self):
        if self._clip_model is None:
            self._clip_model = CLIPModel.from_pretrained(self.config["clip_model"])
        return self._clip_model

    @property
    def clip_processor(self):
        if self._clip_processor is None:
            self._clip_processor = CLIPProcessor.from_pretrained(
                self.config["clip_model"]
            )
        return self._clip_processor

    def load_task(self, task_name_or_config: str | DictConfig):
        if isinstance(task_name_or_config, str):
            task_config = self.get_task_config(task_name_or_config)
        else:
            task_config = task_name_or_config
        task_config = self.prepare_task_config(task_config)

        # load the task from the configuration
        task = CLIPImageClassificationTask(task_config)
        task._fabric = self._fabric
        task._taskpool = self
        task._clip_processor = self.clip_processor

        return task

    def evaluate(self, model: CLIPVisionModel, name=None):
        """
        Evaluate the model on the image classification task.

        Args:
            model (CLIPVisionModel): The vision model to evaluate.

        Returns:
            dict: A dictionary containing the evaluation results for each task.
        """
        # if the fabric is not set, and we have a GPU, create a fabric instance
        if self._fabric is None and torch.cuda.is_available():
            self._fabric = L.Fabric(devices=1)
            self._fabric.launch()

        # CLIPVisionModel works the same with CLIPVisionTransformer, so we can use it directly
        self.clip_model.vision_model = model
        report = {}
        training_params, all_params = count_parameters(model)
        report["model_info"] = {
            "trainable_params": training_params,
            "all_params": all_params,
            "trainable_percentage": training_params / all_params,
        }
        if name is not None:
            report["model_info"]["name"] = name
        for task_name in tqdm(self.task_names, desc="Evaluating tasks"):
            task = self.load_task(task_name)
            result = task.evaluate(self.clip_model)
            report[task_name] = result

        # calculate the average accuracy and loss
        if "average" not in report:
            report["average"] = {}
            accuracies = [
                value["accuracy"]
                for key, value in report.items()
                if "accuracy" in value
            ]
            if len(accuracies) > 0:
                average_accuracy = sum(accuracies) / len(accuracies)
                report["average"]["accuracy"] = average_accuracy
            losses = [value["loss"] for key, value in report.items() if "loss" in value]
            if len(losses) > 0:
                average_loss = sum(losses) / len(losses)
                report["average"]["loss"] = average_loss

        log.info(f"Results for taskpool {self.config.name}: {report}")
        if self._fabric.is_global_zero and len(self._fabric._loggers) > 0:
            with open(
                os.path.join(self._fabric.logger.log_dir, "report.json"), "w"
            ) as fp:
                json.dump(report, fp)
        return report



from typing import Tuple

import torch
from torch import Tensor, nn

from fusion_bench.utils.parameters import state_dict_to_vector
from fusion_bench.utils.state_dict_arithmetic import state_dict_sub


def _svd(w: Tensor, full_matrices=True) -> Tuple[Tensor, Tensor, Tensor]:
    """
    Perform Singular Value Decomposition (SVD) on a tensor.

    Args:
        w (Tensor): The input tensor.
        full_matrices (bool): Whether to compute the full-sized U and V matrices.

    Returns:
        Tuple[Tensor, Tensor, Tensor]: The U, S, and V matrices from SVD.
    """
    u, s, vh = torch.linalg.svd(
        w, full_matrices=full_matrices, driver="gesvd" if w.is_cuda else None
    )
    v = vh.T
    return u, s, v


def svd(
    w: Tensor, full_matrices=True, accelerator=None
) -> Tuple[Tensor, Tensor, Tensor]:
    """
    Perform SVD on a tensor, optionally using a specified accelerator.

    Args:
        w (Tensor): The input tensor.
        full_matrices (bool): Whether to compute the full-sized U and V matrices.
        accelerator (str): The device to perform the computation on.

    Returns:
        Tuple[Tensor, Tensor, Tensor]: The U, S, and V matrices from SVD.
    """
    if accelerator is None:
        return _svd(w, full_matrices=full_matrices)
    original_device = w.device
    w = w.to(accelerator)
    u, s, v = _svd(w)
    return u.to(original_device), s.to(original_device), v.to(original_device)


def frobenius_inner_product(w1: Tensor, w2: Tensor) -> Tensor:
    return torch.trace(w1.T @ w2)


def is_leaf_module(module: nn.Module) -> bool:
    return len(list(module.children())) == 0


def get_task_vector_norm(model: nn.Module, pretrained_model: nn.Module) -> Tensor:
    """
    Get the vector norm of the task model.

    Args:
        model (nn.Module): The task model.
        pretrained_model (nn.Module): The pretrained model.

    Returns:
        Tensor: The vector norm of the task model.
    """
    return torch.linalg.norm(
        state_dict_to_vector(
            state_dict_sub(model.state_dict(), pretrained_model.state_dict())
        )
    )



import logging
from pathlib import Path

import lightning as L
from omegaconf import DictConfig
from torch import nn
from torch.utils.data import DataLoader

from fusion_bench.compat.taskpool.base_pool import TaskPool
from fusion_bench.dataset.nyuv2 import NYUv2
from fusion_bench.models.nyuv2.lightning_module import NYUv2MTLModule
from fusion_bench.models.nyuv2.resnet_dilated import ResnetDilated

log = logging.getLogger(__name__)


class NYUv2TaskPool(TaskPool):
    """Task pool for multi-task learning evaluation on the NYUv2 dataset.

    This task pool provides evaluation capabilities for multi-task learning models
    on the NYU Depth V2 (NYUv2) dataset, which is a popular benchmark for indoor
    scene understanding. The dataset supports multiple computer vision tasks
    including semantic segmentation, depth estimation, and surface normal prediction.

    The task pool is designed to work with encoder-decoder architectures where
    a shared encoder processes input images and task-specific decoders generate
    predictions for different tasks. It integrates with PyTorch Lightning for
    streamlined training and evaluation workflows.

    Supported Tasks:
        - Semantic segmentation
        - Depth estimation
        - Surface normal prediction
    """

    _trainer: L.Trainer = None

    def __init__(self, taskpool_config: DictConfig):
        """Initialize the NYUv2 task pool with configuration settings.

        Args:
            taskpool_config: Configuration object containing all necessary
                parameters for the task pool, including:
                - data_dir: Path to the directory containing NYUv2 dataset
                - tasks: List of tasks to evaluate (e.g., ["semantic", "depth"])
                - batch_size: Batch size for evaluation data loader
                - num_workers: Number of worker processes for data loading
        """
        self.config = taskpool_config

    def load_datasets(self):
        log.info("Loading NYUv2 dataset")
        data_path = str(Path(self.config.data_dir) / "nyuv2")

        train_dataset = NYUv2(root=data_path, train=True)
        val_dataset = NYUv2(root=data_path, train=False)
        return train_dataset, val_dataset

    @property
    def trainer(self):
        if self._trainer is None:
            self._trainer = L.Trainer(devices=1)
        return self._trainer

    def get_decoders(self):
        from fusion_bench.modelpool.nyuv2_modelpool import NYUv2ModelPool

        modelpool: NYUv2ModelPool = self._program.modelpool
        decoders = nn.ModuleDict()
        for task in self.config.tasks:
            decoders[task] = modelpool.load_model(task, encoder_only=False).decoders[
                task
            ]
        return decoders

    def evaluate(self, encoder: ResnetDilated):
        model = NYUv2MTLModule(
            encoder,
            self.get_decoders(),
            tasks=self.config.tasks,
            task_weights=[1] * len(self.config.tasks),
        )
        _, val_dataset = self.load_datasets()
        val_loader = DataLoader(
            val_dataset,
            batch_size=self.config.batch_size,
            shuffle=False,
            num_workers=self.config.num_workers,
        )
        report = self.trainer.validate(model, val_loader)
        if isinstance(report, list) and len(report) == 1:
            report = report[0]
        return report



# Copyright (c) 2020, 2021 The HuggingFace Team
# Copyright (c) 2021 Philip May, Deutsche Telekom AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lazy-Imports module.

This is code taken from the `HuggingFace team <https://huggingface.co/>`__.
Many thanks to HuggingFace for
`your consent <https://github.com/huggingface/transformers/issues/12861#issuecomment-886712209>`__
to publish it as a standalone package.
"""

import importlib
import os
from types import ModuleType
from typing import Any, Dict, List, Optional, Set, Union


class LazyImporter(ModuleType):
    """Lazy importer for modules and their components.

    This class allows for lazy importing of modules, meaning modules are only
    imported when they are actually accessed. This can help reduce startup
    time and memory usage for large packages with many optional dependencies.

    Attributes:
        _modules: Set of module names available for import.
        _class_to_module: Mapping from class/function names to their module names.
        _objects: Dictionary of extra objects to include in the module.
        _name: Name of the module.
        _import_structure: Dictionary mapping module names to lists of their exports.
    """

    # Very heavily inspired by optuna.integration._IntegrationModule
    # https://github.com/optuna/optuna/blob/master/optuna/integration/__init__.py
    def __init__(
        self,
        name: str,
        module_file: str,
        import_structure: Dict[str, List[str]],
        extra_objects: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Initialize the LazyImporter.

        Args:
            name: The name of the module.
            module_file: Path to the module file.
            import_structure: Dictionary mapping module names to lists of their exports.
            extra_objects: Optional dictionary of extra objects to include.
        """
        super().__init__(name)
        self._modules: Set[str] = set(import_structure.keys())
        self._class_to_module: Dict[str, str] = {}
        for key, values in import_structure.items():
            for value in values:
                self._class_to_module[value] = key
        # Needed for autocompletion in an IDE
        self.__all__: List[str] = list(import_structure.keys()) + sum(
            import_structure.values(), []
        )
        self.__file__ = module_file
        self.__path__ = [os.path.dirname(module_file)]
        self._objects: Dict[str, Any] = {} if extra_objects is None else extra_objects
        self._name = name
        self._import_structure = import_structure

    # Needed for autocompletion in an IDE
    def __dir__(self) -> List[str]:
        """Return list of available attributes for autocompletion.

        Returns:
            List of all available attribute names.
        """
        return super().__dir__() + self.__all__

    def __getattr__(self, name: str) -> Any:
        """Get attribute lazily, importing the module if necessary.

        Args:
            name: The name of the attribute to retrieve.

        Returns:
            The requested attribute.

        Raises:
            AttributeError: If the attribute is not found in any module.
        """
        if name in self._objects:
            return self._objects[name]
        if name in self._modules:
            value = self._get_module(name)
        elif name in self._class_to_module:
            module = self._get_module(self._class_to_module[name])
            value = getattr(module, name)
        else:
            raise AttributeError(f"module {self.__name__} has no attribute {name}")

        setattr(self, name, value)
        return value

    def _get_module(self, module_name: str) -> ModuleType:
        """Import and return the specified module.

        Args:
            module_name: Name of the module to import.

        Returns:
            The imported module.
        """
        return importlib.import_module("." + module_name, self.__name__)

    def __reduce__(self) -> tuple:
        """Support for pickling the LazyImporter.

        Returns:
            Tuple containing the class and arguments needed to reconstruct the object.
        """
        return (self.__class__, (self._name, self.__file__, self._import_structure))


class LazyPyModule(ModuleType):
    """Module wrapper for lazy import.

    Adapted from Optuna: https://github.com/optuna/optuna/blob/1f92d496b0c4656645384e31539e4ee74992ff55/optuna/__init__.py

    This class wraps specified module and lazily import it when they are actually accessed.
    This can help reduce startup time and memory usage by deferring module imports
    until they are needed.

    Args:
        name: Name of module to apply lazy import.

    Attributes:
        _name: The name of the module to be lazily imported.
    """

    def __init__(self, name: str) -> None:
        """Initialize the LazyPyModule.

        Args:
            name: The name of the module to be lazily imported.
        """
        super().__init__(name)
        self._name: str = name

    def _load(self) -> ModuleType:
        """Load the actual module and update this object's dictionary.

        Returns:
            The loaded module.
        """
        module = importlib.import_module(self._name)
        self.__dict__.update(module.__dict__)
        return module

    def __getattr__(self, item: str) -> Any:
        """Get attribute from the lazily loaded module.

        Args:
            item: The name of the attribute to retrieve.

        Returns:
            The requested attribute from the loaded module.
        """
        return getattr(self._load(), item)



import logging
import os
from copy import deepcopy
from functools import cache
from typing import Dict, List, cast

import lightning as L
import torch
from omegaconf import DictConfig
from torch import Tensor, nn
from torch.nn.modules import Module
from torch.utils.data import DataLoader
from tqdm.autonotebook import tqdm
from transformers import GPT2ForSequenceClassification, GPT2Model
from transformers.data import default_data_collator
from transformers.models.gpt2.modeling_gpt2 import Conv1D

from fusion_bench.mixins import LightningFabricMixin, auto_register_config
from fusion_bench.modelpool import GPT2ForSequenceClassificationPool
from fusion_bench.utils import timeit_context

from .fisher_merging import FisherMergingAlgorithm, get_param_squared_gradients


@auto_register_config
class FisherMergingAlgorithmForGPT2(
    LightningFabricMixin,
    FisherMergingAlgorithm,
):
    """
    Implements the Fisher Merging Algorithm for GPT-2 models on text classification tasks.

    This class extends the FisherMergingAlgorithm to handle GPT-2 models specifically.
    It supports caching, batch processing, and multi-worker data loading.

    Attributes:
        classifiers (dict): A dictionary to store classifiers for each model.
        modelpool (HuggingFaceGPT2ClassificationPool): The model pool containing the GPT-2 models.
        cache_dir (str): Directory to cache data.
        batch_size (int): Batch size for data loading.
        num_workers (int): Number of workers for data loading.
    """

    classifiers = {}
    modelpool: GPT2ForSequenceClassificationPool = None

    def __init__(
        self,
        cache_dir: str,
        batch_size: int,
        num_workers: int,
        **kwargs,
    ):
        """
        Initialize the FisherMergingAlgorithmForGPT2 with the given configuration.

        Args:
            cache_dir (str): Directory to cache data.
            batch_size (int): Batch size for data loading.
            num_workers (int): Number of workers for data loading.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)

    def on_fisher_merging_start(self):
        """
        Setup the classifiers for each model in the model pool before starting the Fisher merging process.
        """
        for model_name in self.modelpool.model_names:
            classifier = cast(
                GPT2ForSequenceClassification,
                self.modelpool.load_classifier(model_name),
            ).requires_grad_(False)
            classifier.transformer = None
            classifier = classifier.to(self.fabric.device)
            self.classifiers[model_name] = classifier

    def compute_logits(self, module: GPT2Model, batch, task: str) -> Tensor:
        """
        Compute the logits for the given batch and task.

        Args:
            module (GPT2Model): The GPT-2 model module.
            batch (dict): The input batch.
            task (str): The name of the task.

        Returns:
            Tensor: The computed logits.
        """
        self.classifiers[task].transformer = module
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]

        outputs = self.classifiers[task](input_ids, attention_mask=attention_mask)
        logits = outputs.logits
        assert logits.dim() == 2
        return logits

    def get_fisher_weights(
        self,
        model_name: str,
        model: Module,
        train_dataset,
        param_names_to_merge: List[str],
    ) -> Dict[str, Tensor]:
        """
        Compute the Fisher weights for the given model and training dataset.

        Args:
            model_name (str): The name of the model.
            model (Module): The model module.
            train_dataset: The training dataset.
            param_names_to_merge (List[str]): List of parameter names to merge.

        Returns:
            Dict[str, Tensor]: The computed Fisher weights for each parameter.
        """
        # setup dataloader
        train_dataloader = DataLoader(
            train_dataset,
            batch_size=self.config.batch_size,
            shuffle=True,
            collate_fn=default_data_collator,
            num_workers=self.config.num_workers,
            pin_memory=True,
        )
        train_dataloader = self.fabric.setup_dataloaders(train_dataloader)
        model = self.fabric.setup(model)
        num_fisher_examples = self.config.num_fisher_examples
        if num_fisher_examples % train_dataloader.batch_size != 0:
            print(
                f"warning: the number of examples for computing fisher cannot be fully divided by the batch size for model, "
                "which may lead to a slightly different number of the actually used examples."
            )
        num_computed_examples = 0
        batches_fisher_weights_list = []
        for step, batch in tqdm(
            enumerate(train_dataloader),
            desc=f"computing fisher weights",
            total=num_fisher_examples // train_dataloader.batch_size,
        ):
            if num_computed_examples >= num_fisher_examples:
                break
            logits = self.compute_logits(model, batch, model_name)
            # Tensor, shape (batch_size, num_label_classes)

            # compute fisher weights for classifxication task
            # use detach() to detach from the computation graph
            # Tensor, shape (batch_size, num_label_classes)
            labels_probabilities = torch.softmax(logits, dim=-1).detach()
            labels_log_probabilities = torch.log_softmax(logits, dim=-1)
            # sqrt labels_probabilities, since torch.sqrt(labels_probabilities) would be squared in the following squared gradients
            labels_expectations = (
                torch.sqrt(labels_probabilities) * labels_log_probabilities
            )
            # sum over label classes and batch dimension
            sum_labels_expectations = labels_expectations.sum(dim=-1).sum(dim=0)
            model.zero_grad()
            sum_labels_expectations.backward()
            # dict, fisher weights of a batch
            batch_fisher_weights = get_param_squared_gradients(
                model=model, param_names_to_merge=param_names_to_merge
            )

            # move fisher weights to cpu to save GPU memory
            for key, weights in batch_fisher_weights.items():
                batch_fisher_weights[key] = weights.detach().cpu()

            batches_fisher_weights_list.append(batch_fisher_weights)
            num_computed_examples += batch["input_ids"].size(0)

        model_to_merge_fisher_weights = {}
        for batch_fisher_weights in batches_fisher_weights_list:
            for key in batch_fisher_weights:
                if key not in model_to_merge_fisher_weights:
                    model_to_merge_fisher_weights[key] = batch_fisher_weights[key]
                else:
                    model_to_merge_fisher_weights[key] += batch_fisher_weights[key]

        # mean over batches
        for key in model_to_merge_fisher_weights:
            model_to_merge_fisher_weights[key] /= num_computed_examples
            model_to_merge_fisher_weights[key] = (
                model_to_merge_fisher_weights[key].detach().cpu()
            )
        return model_to_merge_fisher_weights



import torch
from torch import Tensor, nn


def segmentation_loss(pred: Tensor, gt: Tensor):
    return nn.functional.cross_entropy(pred, gt.long(), ignore_index=-1)


def depth_loss(pred: Tensor, gt: Tensor):
    binary_mask = (torch.sum(gt, dim=1) != 0).float().unsqueeze(1).to(pred.device)
    loss = torch.sum(torch.abs(pred - gt) * binary_mask) / torch.nonzero(
        binary_mask, as_tuple=False
    ).size(0)
    return loss


def normal_loss(pred: Tensor, gt: Tensor):
    # gt has been normalized on the NYUv2 dataset
    pred = pred / torch.norm(pred, p=2, dim=1, keepdim=True)
    binary_mask = (torch.sum(gt, dim=1) != 0).float().unsqueeze(1).to(pred.device)
    loss = 1 - torch.sum((pred * gt) * binary_mask) / torch.nonzero(
        binary_mask, as_tuple=False
    ).size(0)
    return loss


loss_fn = {
    "segmentation": segmentation_loss,
    "depth": depth_loss,
    "normal": normal_loss,
}



import logging
from typing import Mapping, Optional

from lightning_utilities.core.rank_zero import rank_prefixed_message, rank_zero_only

__all__ = [
    "RankedLogger",
    "RankZeroLogger",
    "get_rankzero_logger",
]


class RankedLogger(logging.LoggerAdapter):
    """A multi-GPU-friendly python command line logger."""

    def __init__(
        self,
        name: str = __name__,
        rank_zero_only: bool = False,
        extra: Optional[Mapping[str, object]] = None,
    ) -> None:
        """Initializes a multi-GPU-friendly python command line logger that logs on all processes
        with their rank prefixed in the log message.

        :param name: The name of the logger. Default is ``__name__``.
        :param rank_zero_only: Whether to force all logs to only occur on the rank zero process. Default is `False`.
        :param extra: (Optional) A dict-like object which provides contextual information. See `logging.LoggerAdapter`.
        """
        logger = logging.getLogger(name)
        super().__init__(logger=logger, extra=extra)
        self.rank_zero_only = rank_zero_only

    def log(
        self, level: int, msg: str, rank: Optional[int] = None, *args, **kwargs
    ) -> None:
        """Delegate a log call to the underlying logger, after prefixing its message with the rank
        of the process it's being logged from. If `'rank'` is provided, then the log will only
        occur on that rank/process.

        :param level: The level to log at. Look at `logging.__init__.py` for more information.
        :param msg: The message to log.
        :param rank: The rank to log at.
        :param args: Additional args to pass to the underlying logging function.
        :param kwargs: Any additional keyword args to pass to the underlying logging function.
        """
        if self.isEnabledFor(level):
            msg, kwargs = self.process(msg, kwargs)
            current_rank = getattr(rank_zero_only, "rank", None)
            if current_rank is None:
                raise RuntimeError(
                    "The `rank_zero_only.rank` needs to be set before use"
                )
            msg = rank_prefixed_message(msg, current_rank)
            if self.rank_zero_only:
                if current_rank == 0:
                    self.logger.log(level, msg, *args, **kwargs)
            else:
                if rank is None:
                    self.logger.log(level, msg, *args, **kwargs)
                elif current_rank == rank:
                    self.logger.log(level, msg, *args, **kwargs)


class RankZeroLogger(logging.Logger):
    """A logger that logs only on rank zero and works just like logging.Logger"""

    @rank_zero_only
    def _log(self, *args, **kwargs):
        if "stacklevel" in kwargs:
            kwargs["stacklevel"] += 1
        else:
            kwargs["stacklevel"] = 2
        return super()._log(*args, **kwargs)

    def is_global_zero(self):
        return rank_zero_only.rank == 0


RankZeroLogger.manager = logging.Manager(RankZeroLogger.root)
RankZeroLogger.manager.setLoggerClass(RankZeroLogger)


def get_rankzero_logger(name=None):
    """
    Return a logger with the specified name, creating it if necessary.

    If no name is specified, return the root logger.
    """
    if not name or isinstance(name, str) and name == logging.root.name:
        return logging.root
    return RankZeroLogger.manager.getLogger(name)



"""
Dummy task pool implementation for debugging and testing purposes.

This module provides a minimal task pool implementation that can be used for
debugging model fusion workflows, testing infrastructure, and validating model
architectures without running expensive evaluation procedures. It's particularly
useful during development and prototyping phases.
"""

from typing import Optional

from lightning.pytorch.utilities import rank_zero_only
from torch import nn

from fusion_bench.models.separate_io import separate_save
from fusion_bench.taskpool.base_pool import BaseTaskPool
from fusion_bench.utils import timeit_context
from fusion_bench.utils.parameters import count_parameters, print_parameters


def get_model_summary(model: nn.Module) -> dict:
    """Generate a comprehensive summary report for a PyTorch model.

    Analyzes the given model to extract key information about its architecture,
    parameter count, and training characteristics. This function is useful for
    model introspection and comparative analysis during model fusion workflows.

    The summary includes both trainable and total parameter counts, which helps
    in understanding model complexity and memory requirements. The trainable
    percentage is particularly useful for identifying models with frozen layers
    or parameter-efficient fine-tuning setups.

    Args:
        model: The PyTorch model to analyze. Can be any nn.Module instance
            including complex models, fusion models, or pre-trained models.

    Returns:
        dict: A structured report containing model information:
            - model_info: Dictionary with parameter statistics
                - trainable_params: Number of trainable parameters
                - all_params: Total number of parameters (trainable + frozen)
                - trainable_percentage: Ratio of trainable to total parameters

    Example:
        ```python
        >>> model = MyModel()
        >>> summary = get_model_summary(model)
        >>> print(summary)
        {
            "model_info": {
                "trainable_params": 1234567,
                "all_params": 1234567,
                "trainable_percentage": 1.0
            }
        }
        ```
    """
    report = {}
    training_params, all_params = count_parameters(model)
    report["model_info"] = {
        "trainable_params": training_params,
        "all_params": all_params,
        "trainable_percentage": training_params / all_params,
    }
    return report


class DummyTaskPool(BaseTaskPool):
    """A lightweight task pool implementation for debugging and development workflows.

    This dummy task pool provides a minimal evaluation interface that focuses on
    model introspection rather than task-specific performance evaluation. It's
    designed for development scenarios where you need to test model fusion
    pipelines, validate architectures, or debug workflows without the overhead
    of running actual evaluation tasks.

    The task pool is particularly useful when:
        - You want to verify model fusion works correctly
        - You need to check parameter counts after fusion
        - You're developing new fusion algorithms
        - You want to test infrastructure without expensive evaluations

    Example:
        ```python
        >>> taskpool = DummyTaskPool(model_save_path="/tmp/fused_model")
        >>> results = taskpool.evaluate(fused_model)
        >>> print(f"Model has {results['model_info']['trainable_params']} parameters")
        ```
    """

    def __init__(self, model_save_path: Optional[str] = None, **kwargs):
        """Initialize the dummy task pool with optional model saving capability.

        Args:
            model_save_path: Optional path where the evaluated model should be saved.
                If provided, the model will be serialized and saved to this location
                after evaluation using the separate_save utility. If None, no model
                saving will be performed.

        Example:
            ```python
            >>> # Create taskpool without saving
            >>> taskpool = DummyTaskPool()

            >>> # Create taskpool with model saving
            >>> taskpool = DummyTaskPool(model_save_path="/path/to/save/model.pth")
            ```
        """
        super().__init__(**kwargs)
        self.model_save_path = model_save_path

    def evaluate(self, model):
        """Perform lightweight evaluation and analysis of the given model.

        This method provides a minimal evaluation that focuses on model introspection
        rather than task-specific performance metrics. It performs parameter analysis,
        optionally saves the model, and returns a summary report.

        The evaluation process includes:
        1. Printing human-readable parameter information (rank-zero only)
        2. Optionally saving the model if a save path was configured
        3. Generating and returning a model summary report

        Args:
            model: The model to evaluate. Can be any PyTorch nn.Module including
                fusion models, pre-trained models, or custom architectures.

        Returns:
            dict: A model summary report containing parameter statistics and
                architecture information. See get_model_summary() for detailed
                format specification.

        Example:
            ```python
            >>> taskpool = DummyTaskPool(model_save_path="/tmp/model.pth")
            >>> model = torch.nn.Linear(10, 5)
            >>> results = taskpool.evaluate(model)
            >>> print(f"Trainable params: {results['model_info']['trainable_params']}")
            ```
        """
        if rank_zero_only.rank == 0:
            print_parameters(model, is_human_readable=True)

            if self.model_save_path is not None:
                with timeit_context(f"Saving the model to {self.model_save_path}"):
                    separate_save(model, self.model_save_path)

        return get_model_summary(model)



from transformers import AutoConfig, AutoModel, AutoModelForCausalLM

from .configuration_smile_llama import SmileLlamaConfig
from .modeling_smile_llama import SmileLlamaForCausalLM, SmileLlamaModel

AutoConfig.register("smile_llama", SmileLlamaConfig)
AutoModel.register(SmileLlamaConfig, SmileLlamaModel)
AutoModelForCausalLM.register(SmileLlamaConfig, SmileLlamaForCausalLM)



import torch


class CacheDataset(torch.utils.data.Dataset):
    def __init__(self):
        self.alphas = []  # logits
        self.Xs = []  # input hidden states
        self.Zs = []  # output hidden states
        self.prepared = False

    def __len__(self):
        if not self.prepared:
            self.prepare_for_loader()
        return len(self.alphas)

    def __getitem__(self, index):
        if not self.prepared:
            self.prepare_for_loader()
        if isinstance(index, list):
            return [(self.alphas[idx], self.Xs[idx], self.Zs[idx]) for idx in index]
        elif isinstance(index, int):
            return self.alphas[index], self.Xs[index], self.Zs[index]

    def append(self, alpha=None, X=None, Z=None):
        if alpha is not None:
            self.alphas.append(alpha.detach().to("cpu", non_blocking=True))
        if X is not None:
            self.Xs.append(X.detach().to("cpu", non_blocking=True))
        if Z is not None:
            self.Zs.append(Z.detach().to("cpu", non_blocking=True))
        self.prepared = False

    def prepare_for_loader(self):
        if self.prepared:
            return
        self.prepared = True
        self.alphas = torch.concat(self.alphas)
        self.Xs = torch.concat(self.Xs)
        self.Zs = torch.concat(self.Zs)
        assert len(self.Xs) == len(self.Zs)
